
from kivymd.bottomsheet import MDListBottomSheet, MDGridBottomSheet
from kivymd.button import MDIconButton
from kivymd.date_picker import MDDatePicker
from kivymd.dialog import MDDialog
from kivymd.label import MDLabel
from kivymd.list import ILeftBody, ILeftBodyTouch, \
			IRightBodyTouch, BaseListItem
from kivymd.material_resources import DEVICE_TYPE
from kivymd.navigationdrawer import MDNavigationDrawer, NavigationDrawerHeaderBase
from kivymd.navigationdrawer import NavigationLayout
from kivymd.navigationdrawer import NavigationDrawerDivider
from kivymd.navigationdrawer import NavigationDrawerToolbar
from kivymd.navigationdrawer import NavigationDrawerSubheader
from kivymd.selectioncontrols import MDSwitch
from kivymd.selectioncontrols import MDCheckbox
from kivymd.snackbar import Snackbar
from kivymd.theming import ThemeManager
from kivymd.time_picker import MDTimePicker
from kivymd.toolbar import Toolbar
from kivymd.list import MDList
from kivymd.textfields import MDTextField
from kivymd.spinner import MDSpinner
from kivymd.card import MDCard
from kivymd.card import MDSeparator
from kivymd.menu import MDDropdownMenu
from kivymd.grid import SmartTile
from kivymd.slider import MDSlider
from kivymd.tabs import MDTabbedPanel, MDTab, MDBottomNavigation, \
	MDBottomNavigationItem
from kivymd.progressbar import MDProgressBar
from kivymd.accordion import MDAccordion, MDAccordionItem
from kivymd.theme_picker import MDThemePicker

from kivyblocks.blocks import registerWidget

class TabbedPannel(MDTabbedPanel):
	def __init__(self,**options):
		
def kivyMDPlugin():
	registerWidget('MDSlider',MDSlider)
	registerWidget('MDTabbedPanel',MDTabbedPanel)
	registerWidget('MDTab',MDTab)
	registerWidget('MDBottomNavigation',MDBottomNavigation)
	registerWidget('MDBottomNavigationItem',MDBottomNavigationItem)
	registerWidget('MDProgressBar',MDProgressBar)
	registerWidget('MDAccordion',MDAccordion)
	registerWidget('MDAccordionItem',MDAccordionItem)
	registerWidget('MDThemePicker',MDThemePicker)
	registerWidget('MDListBottomSheet',MDListBottomSheet)
	registerWidget('MDGridBottomSheet',MDGridBottomSheet)
	registerWidget('MDIconButton', MDIconButton)
	registerWidget('MDDatePicker', MDDatePicker)
	registerWidget('MDDialog', MDDialog)
	registerWidget('MDLabel', MDLabel)
	registerWidget('ILeftBody',ILeftBody)
	registerWidget('ILeftBodyTouch',ILeftBodyTouch)
	registerWidget('IRightBodyTouch',IRightBodyTouch)
	registerWidget('BaseListItem',BaseListItem)
	registerWidget('MDNavigationDrawer',MDNavigationDrawer)
	registerWidget('NavigationDrawerHeaderBase',NavigationDrawerHeaderBase)
	registerWidget('MDCheckbox',MDCheckbox)
	registerWidget('Snackbar',Snackbar)
	registerWidget('ThemeManager',ThemeManager)
	registerWidget('MDTimePicker',MDTimePicker)
	registerWidget('Toolbar',Toolbar)
	registerWidget('NavigationLayout',NavigationLayout)
	registerWidget('NavigationDrawerDivider',NavigationDrawerDivider)
	registerWidget('NavigationDrawerToolbar',NavigationDrawerToolbar)
	registerWidget('NavigationDrawerSubheader',NavigationDrawerSubheader)
	registerWidget('MDSwitch',MDSwitch)
	registerWidget('MDList',MDList)
	registerWidget('MDTextField',MDTextField)
	registerWidget('MDSpinner',MDSpinner)
	registerWidget('MDCard',MDCard)
	registerWidget('MDSeparator',MDSeparator)
	registerWidget('SmartTile',SmartTile)




from kivy.event import EventDispatcher

class DataLoader(EventDispatcher):
	def __init__(self,**kw):
		self.register_event_type('on_loaded')
		self.register_event_type('on_loaderror')
		super(DataLoader,self).__init__(**kw)
	
	def loadData(self):
		pass
		
	def dataLoaded(self,d):
		self.dispatch('on_loaded',d)
		
	def loadError(self,e):
		self.dispatch('on_loaderror',e)

	def on_loaded(self,d):
		pass #print('on_loaded,data=',d)

	def on_loaderror(self,a,e):
		pass #print('error:',e)



import requests

headers = {
   'User-Agent': "Mozilla/5.0 (Windows NT 6.1; Win64; x64) "
                 "AppleWebKit/537.36 (KHTML, like Gecko) "
                 "Chrome/72.0.3626.109 Safari/537.36",
}

def read_m3u8(m3u8Url):
	baseurl = '/'.join(m3u8Url.split('/')[:-1])
	m3u8_res = requests.get(m3u8Url,headers)
	if m3u8_res.status_code != 200:
		print('URL不能正常访问', m3u8_res.status_code)
	m3u8_content = m3u8_res.content.decode('utf8')
	print(m3u8_content)
	'''获取m3u8中的下载连接'''
	media_url_list = []
	lines_list = m3u8_content.strip().split('\r\n')
	if len(lines_list) < 3:
		lines_list = m3u8_content.strip().split('\n')
	if '#EXTM3U' not in m3u8_content:
		raise BaseException('非M3U8连接')
	for index,line in enumerate(lines_list):
		# print(index,line)
		if '#EXTINF' in line:
			media_url_list.append('%s/%s' %(baseurl,lines_list[index+1]))
	return media_url_list

if __name__ == '__main__':
	vfile = read_m3u8("http://pili-live-hdl.qhmywl.com/dsdtv/6d735caeaf70f8901aa00f86aefc4dde.m3u8")
	print(vfile)



from kivy.logger import Logger
from kivy.properties import NumericProperty, ListProperty, StringProperty
from kivy.graphics import Color, Rectangle, RoundedRectangle
from kivy.uix.textinput import TextInput
from kivy.uix.button import Button
from kivy.uix.label import Label
from kivyblocks.utils import CSize

from .colorcalc import *

csses = {
	"default":{
		"bgcolor":[0.35,0.35,0.35,1],
		"fgcolor":[0.85,0.85,0.85,1]
	}
}

csskeys=[
	"height_nm",
	"width_nm",
	"height_c",
	"width_c",
	"fgcolor_s",
	"bgcolor_s",
	"bgcolor",
	"fgcolor",
	"csscls",
	"radius",
	"height",
	"width",
]

def register_css(cssname, cssdic):
	dic = {k:v for k,v in cssdic.items() if k in csskeys }
	csses.update({cssname:dic})

def get_css(cssname,default='default'):
	return csses.get(cssname, csses.get(default))

class WidgetCSS(object):
	height_nm = NumericProperty(None)
	width_nm = NumericProperty(None)
	height_c = NumericProperty(None)
	width_c = NumericProperty(None)
	fgcolor_s = StringProperty(None)
	bgcolor_s = StringProperty(None)
	bgcolor = ListProperty(None)
	fgcolor = ListProperty(None)
	csscls = StringProperty("default")
	radius = ListProperty(None)
	background_rect = None
	bg_func = Rectangle

	def on_canvas(self, o, s):
		#Logger.info('WidgetCSS:on_canvas():%s',self.__class__.__name__)
		self.set_background_color()

	def on_size(self, o, s):
		try:
			super(WidgetCSS, self).on_size(o, s)
		except:
			pass
		self.set_background_color()

	def on_pos(self, o, p):
		self.set_background_color()

	def on_height_nm(self, o, v):	
		if not self.height_nm:
			return
		if self.height_nm <= 1:
			self.size_hint_y = self.height_nm
		else:
			self.size_hint_y = None
			self.height = height_nm

	def on_width_nm(self, o, v):
		if not self.width_nm:
			return
		if self.width_nm <= 1:
			self.size_hint_x = self.width_nm
		else:
			self.size_hint_x = None
			self.width = self.width_nm

	def on_csscls(self, o, csscls):
		self.set_css(self.csscls)

	def set_css(self, css_str):
		css = {}
		for css_name in css_str.split(' '):
			css.update(get_css(css_name))
		for k,v in css.items():
			setattr(self,k,v)

	def on_height_c(self, o, c):
		if not self.height_c:
			return
		self.size_hint_y = None
		self.height = CSize(self.height_c)

	def on_width_c(self, o, c):
		if not self.width_c:
			return
		self.size_hint_x = None
		self.width = CSize(self.width_c)

	def set_child_fgcolor(self, c):
		if not self.fgcolor:
			return
		if isinstance(c, WidgetCSS):
			return 
		if isinstance(c, TextInput):
			c.foreground_color = self.fgcolor
			return
		if isinstance(c, Label):
			c.color = self.fgcolor
			return
		for x in c.children:
			self.set_child_fgcolor(x)

	def on_children(self, o, c):
		for c in self.children:
			self.set_child_fgcolor(c)
			
	def on_fgcolor_s(self, o, c):
		if not c:
			return
		self.fgcolor = toArrayColor(c)

	def on_bgcolor_s(self, o, c):
		if not c:
			return
		self.bgcolor = toArrayColor(c)

	def on_fgcolor(self, o, c):
		if not self.fgcolor:
			return
		if isinstance(self, TextInput):
			self.foreground_color = self.fgcolor
		if isinstance(self, Label):
			self.color = self.fgcolor
			return
		for c in self.children:
			self.set_child_fgcolor(c)
		return

	def on_bgcolor(self, o, c):
		if not self.bgcolor:
			return
		if isinstance(self, TextInput):
			self.background_color = self.bgcolor
			return
		if isinstance(self, Button):
			self.background_color = self.bgcolor
			return
		self.set_background_color()

	def on_radius(self, o, r):
		if not self.radius:
			self.bg_func = Rectangle
		else:
			self.bg_func = RoundedRectangle
		self.set_background_color()

	def set_background_color(self, *args):
		if not self.bgcolor:
			return
		if not self.canvas:
			#Logger.info('WidgetCSS: set_bg_color(), canvas is null(%s)',
			#		self.__class__.__name__)
			return

		# self.canvas.before.clear()
		with self.canvas.before:
			Color(*self.bgcolor)
			if self.radius:
				self.background_rect = self.bg_func(pos=self.pos,
							size=self.size,
							radius=self.radius)
			else:
				self.background_rect = self.bg_func(pos=self.pos, 
							size=self.size)




import pyaudio
from kivy.uix.label import Label
from kivy.properties import StringProperty
from kivyblocks.baseWidget import HBox, VBox
import sys
import time

from .audio import Audio

class Recorder(VBox):
	input_id = 
	def __init__(self, **kw):
		super().__init__(**kw)
		self.auido = pyaudio.PyAudio()
		self.input_device_id = 
	def pause(self):
		pass

	def record(self):
		pass

	def stop(self):
		pass

	def play(self):
		pass

	def save(self):
		pass






from kivy.uix.label import Label
from kivy.properties import NumericProperty, StringProperty
from ..baseWidget import FloatInput
from ..utils import *
from .factory import UiFactory, get_value

class FloatView(Label):
	value = NumericProperty(None)
	nullstr = StringProperty('')
	dec = NumericProperty(2)
	def __init__(self, **kw):
		super(FloatView, self).__init__(text=self.nullstr, **kw)
		self.wrap = True
		self.halign = 'right'
		self.valign = 'middle'

	def on_value(self, *args):
		if self.value is None:
			self.text = self.bullstr
		f = '%%.0%df' % self.dec
		self.text = f % self.value
		
	def getValue(self):
		return self.value

	def setValue(self, v):
		self.value = v

	@classmethod
	def builder(self, desc, rec=None):
		v = get_value(desc, rec=rec)
		return FloatView(value=v,
						dec=desc.get('dec', 2),
						nullstr=desc.get('nullstr',''),
						font_size=CSize(1), 
						halign='right', 
						valign='middle'
				)

def build_input_float_widget(desc, rec=None):
	v = get_value(desc, rec=rec)
	kw = desc.get('uiparams', {})
	w = FloatInput(**kw)
	if v is not None:
		w.setValue(v)
	return w
	
UiFactory.register('float', build_input_float_widget, FloatView.builder)




from appPublic.timeUtils import curDateString, monthMaxDay

from kivy.clock import Clock
from kivy.factory import Factory
from kivy.properties import NumericProperty, BooleanProperty, \
			OptionProperty
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label
from kivyblocks.baseWidget import SelectInput, HBox
from kivyblocks.utils import CSize

date_format_info = {
	'yyyy-mm-dd':{
		'ystart':0,
		'mstart':5,
		'dstart':8
	},
	'mm-dd-yyyy':{
		'ystart':6,
		'mstart':0,
		'dstart':3
	},
	'dd-mm-yyyy':{
		'ystart':6,
		'mstart':3,
		'dstart':0
	},
	'yyyymmdd':{
		'ystart':0,
		'mstart':4,
		'dstart':6
	},
	'mmddyyyy':{
		'ystart':4,
		'mstart':0,
		'dstart':2
	},
	'ddmmyyyy':{
		'ystart':4,
		'mstart':2,
		'dstart':0
	}
}
def get_datestr(y, m, d, dateformat):
	def replace_slice(s, v, pos,length):
		x = '%0' + str(length) + 'd'
		sv = x % v
		d = ''
		if pos > 0:
			d = s[0:pos]
		d = d + sv
		d = d + s[pos+length:]
		return d
	if y is None:
		return None

	info = date_format_info.get(dateformat)
	if info is None:
		return None
	ret = dateformat
	ret = replace_slice(ret, y, info['ystart'], 4)
	ret = replace_slice(ret, m, info['mstart'], 4)
	ret = replace_slice(ret, d, info['dstart'], 4)
	return ret

def get_ymd(dstr, dateformat):
	if not dstr:
		return None, None, None

	info = date_format_info.get(dateformat)
	if info is None:
		return None, None, None
	y = int(dstr[info['ystart'], info['ystart']+4])
	m = int(dstr[info['mstart'], info['mstart']+2])
	d = int(dstr[info['dstart'], info['dstart']+2])
	return y, m, d

class YearInput(SelectInput):
	nullable = BooleanProperty(True)
	e_year = NumericProperty(None)
	def __init__(self, show_years=10, **kw):
		super(YearInput, self).__init__(**kw)
		self.show_years = show_years
		self.showed_min = None
		self.showed_max = None
		self.valueField = self.dropdown.valueField
		self.textField = self.dropdown.textField
		self.reopen = False
		self.dropdown.bind(on_dismiss=self.reopen_dropdown)

	def e_year_change(self, y):
		self.e_year = y
		smin = self.e_year - int((self.show_years)/2)
		smax = smin + self.show_years
		self.set_selectable_data(self.years_data(smin, smax))
		
	def years_data(self, smin, smax):
		d = [
			{
				self.valueField:-10000,
				self.textField:'-'
			}
		]
		self.showed_min = smin
		self.showed_max = smax
		while smin < smax:
			d.append({
				self.valueField:smin,
				self.textField:'%4d' % smin
			})
			smin += 1
		if self.nullable:
			d.append({
				None, ''
			})
		d.append({
			self.valueField:10000,
			self.textField:'+'
		})
		return d

	def reopen_dropdown(self, *args):
		if not self.reopen:
			return
		self.dropdown.open(self)

	def dropdown_select(self, o, d):
		if d[0] == -10000:
			smin = self.showed_min - self.show_years
			if smin < 0:
				smin = 0
			smax = smin + self.show_years
			self.set_selectable_data(self.years_data(smin, smax))
			self.reopen = True
			return
		if d[0] == 10000:
			smax = self.showed_max + self.show_years
			if smax > 9999:
				smax = 9999
			smin = smax - self.show_years
			self.set_selectable_data(self.years_data(smin, smax))
			self.reopen = True
			return
		super().dropdown_select(o,d)
		self.reopen = False

	def setValue(self, v):
		if v is None:
			super().setValue('')
			return

		super().setValue(v)
		self.e_year_change(v)

class DateInput(HBox):
	default_date = OptionProperty(None, options=[None, 'today'])
	date_format = OptionProperty('yyyy-mm-dd', options=[k for k in date_format_info.keys() ])
	def __init__(self, allow_copy=True, **kw):
		print('DateInput():kw=', kw)
		kw['size_hint_x'] = None
		kw['width'] = 10
		super(DateInput, self).__init__(**kw)
		self.register_event_type('on_changed')
		self.old_datestr = None
		value = kw.get('value',self.defaultdate())
		y, m, d = get_ymd(value, 'yyyy-mm-dd')
		months_data = []
		days_data = []
		for i in range(12):
			j = i + 1
			months_data.append({
				'text':'%02d' % j,
				'value':j
			})
		for i in range(31):
			j = i + 1
			days_data.append({
				'text':'%02d' % j,
				'value':j
			})
		self.days_data = days_data
		nullable = False
		if value is None:
			nullable = True
		self.yw = YearInput(data=[], nullable=nullable,
						size_hint_x=None, width=3.6)
		self.mw = SelectInput(size_hint_x=None, width=2,
						data=months_data)
		self.dw = SelectInput( size_hint_x=None, width=2,
						data=days_data)
		self.mw.set_selectable_data(months_data)
		self.dw.set_selectable_data(days_data)
		self.yw.setValue(y)
		self.mw.setValue(m)
		self.dw.setValue(d)
		self.add_widget(self.yw)
		self.add_widget(Label(text='-',size_hint_x=None, width=CSize(1)))
		self.add_widget(self.mw)
		self.add_widget(Label(text='-',size_hint_x=None, width=CSize(1)))
		self.add_widget(self.dw)
		self.yw.bind(on_changed=self.data_changed)
		self.mw.bind(on_changed=self.data_changed)
		self.dw.bind(on_changed=self.data_changed)
		self.on_size()

	def defaultdate(self):
		if self.default_date == 'today':
			return curDateString()
		return None

	def on_size(self, *args):
		if not hasattr(self, 'yw'):
			return 
		self.yw.height = self.height
		self.mw.height = self.height
		self.dw.height = self.height

	def str2ymd(self, datestr):
		if datestr is None or datestr == '':
			return None, None, None
		return get_ymd(datestr, self.dateformat)

	def ymd2str(self, y, m, d):
		if y is None:
			return None

		return get_datestr(y, m, d, self.dateformat)

	def data_changed(self, o, d):
		datestr = None
		if y is not None:
			y = self.yw.getValue()
			m = self.mw.getValue()
			d = self.dw.getValue()
			mdays = monthMaxDay(y,m)
			if o == self.yw or o == self.mw:
				data = self.days_data[:mdays]
				self.dw.set_selectable_data(data)
			if d <= mdays and d>0:
				datestr = self.ymd2str(y,m,d)

		if self.old_datestr != datestr:
			self.old_datestr = datestr
			self.dispatch('on_changed', datestr)

	def on_changed(self, *args):
		pass

	def getValue(self):
		y = self.yw.getValue()
		if y is not None:
			m = self.mw.getValue()
			d = self.dw.getValue()
			mdays = monthMaxDay(y,m)
			if d <= mdays and d>0:
				return self.ymd2str(y,m,d)
		return None

	def setValue(self, datestr):
		if datestr == '' or datestr is None:
			datestr = self.defaultdate()
		self.old_value = datestr
		y, m, d = self.str2ymd(datestr)
		self.yw.setValue(y)
		self.mw.setValue(m)
		self.dw.setValue(d)




from kivy.uix.label import Label
from kivy.properties import NumericProperty, StringProperty
from ..baseWidget import IntegerInput
from ..utils import *
from .factory import UiFactory, StringView

class IntView(Label):
	value = NumericProperty(None)
	nullstr = StringProperty('')
	def __init__(self, **kw):
		super(IntView, self).__init__(text='', **kw)
		self.wrap = True
		self.halign = 'right'
		self.valign = 'middle'

	def on_value(self, *args):
		if self.value is not None:
			self.text = str(v)
		else:
			self.text = self.nullstr

	def setValue(self, v):
		self.value = v

	def getValue(self):
		return self.value

	@classmethod
	def builder(self,desc, rec=None):
		v = get_value(desc, rec=rec)
		return IntValue(value=v,
				nullstr=desc.get('nullstr',''),
				font_size=CSize(1), 
				halign='right', 
				valign='middle'
		)

def build_input_int_widget(desc, rec=None):
	v = get_value(desc, rec=rec)
	kw = desc.get('uiparams', {})
	w = IntegerInput(**kw)
	if v is not None:
		w.setValue(v)
	return w
	
UiFactory.register('int',  build_input_int_widget, IntView.builder)




"""
tree description json format:
{
	"url":a http(s) url for get data,
	"checkbox":boolean, show a checkbox before text
	"data":if undefined url,use data to construct the tree
	"params":parameters attached to the http(s) request  via url
	"headers":headers need to set for the http request via url
	"height":widget 's height
	"width":widget's width
}

data structure :
{
	"id":identified field,
	"widgettype":widget type,
	...other data...
	"__children__":[
	]
}
"""
from .scrollwidget import ScrollWidget
from kivy.uix.treeview import TreeView




# -*- coding: utf-8 -*-
# Two implementations. The first is based on color brightness obtained from:-
# https://www.w3.org/TR/AERT#color-contrast
# The second is based on relative luminance calculation for sRGB obtained from:-
# https://www.w3.org/TR/2008/REC-WCAG20-20081211/#relativeluminancedef
# and contrast ratio calculation obtained from:-
# https://www.w3.org/TR/2008/REC-WCAG20-20081211/#contrast-ratiodef
#
# Preliminary testing suggests color brightness more closely matches the
# Material Design spec suggested text colors, but the alternative implementation
# is both newer and the current 'correct' recommendation, so is included here
# as an option.


def _color_brightness(color):
    # Implementation of color brightness method
    brightness = color[0] * 299 + color[1] * 587 + color[2] * 114
    brightness = brightness
    return brightness


def _black_or_white_by_color_brightness(color):
    if _color_brightness(color) >= 500:
        return 'black'
    else:
        return 'white'


def _normalized_channel(color):
    # Implementation of contrast ratio and relative luminance method
    if color <= 0.03928:
        return color / 12.92
    else:
        return ((color + 0.055) / 1.055) ** 2.4


def _luminance(color):
    rg = _normalized_channel(color[0])
    gg = _normalized_channel(color[1])
    bg = _normalized_channel(color[2])
    return 0.2126*rg + 0.7152*gg + 0.0722*bg


def _black_or_white_by_contrast_ratio(color):
    l_color = _luminance(color)
    l_black = 0.0
    l_white = 1.0
    b_contrast = (l_color + 0.05) / (l_black + 0.05)
    w_contrast = (l_white + 0.05) / (l_color + 0.05)
    return 'white' if w_contrast >= b_contrast else 'black'


def get_contrast_text_color(color, use_color_brightness=True):
    if use_color_brightness:
        contrast_color = _black_or_white_by_color_brightness(color)
    else:
        contrast_color = _black_or_white_by_contrast_ratio(color)
    if contrast_color == 'white':
        return 1, 1, 1, 1
    else:
        return 0, 0, 0, 1

if __name__ == '__main__':
    from kivy.utils import get_color_from_hex
    from kivymd.color_definitions import colors, text_colors
    for c in colors.items():
        if c[0] in ['Light', 'Dark']:
            continue
        print("For the {} color palette:".format(c[0]))
        for name, hex_color in c[1].items():
            if hex_color:
                col = get_color_from_hex(hex_color)
                col_bri = get_contrast_text_color(col)
                con_rat = get_contrast_text_color(col, use_color_brightness=False)
                text_color = text_colors[c[0]][name]
                print("   The {} hue gives {} using color brightness, {} using contrast ratio, and {} from the MD spec"
                      .format(name, col_bri, con_rat, text_color))



from appPublic.jsonConfig import getConfig
from kivy.uix.button import ButtonBehavior
from kivy.app import App
from kivy.uix.boxlayout import BoxLayout
from .baseWidget import *
from .objectViewer import ObjectViewer

class ServerImageViewer(ObjectViewer):
	def showObject(self,rec,**kw):
		blocks = App.get_running_app().blocks
		config = getConfig()
		url = '%s/%s/%s' % ( config.uihome,'thumb',rec['id'])
		desc = self.viewer.copy()
		desc['options'].update({
			"size_hint_x":None,
			"width":self.box_width,
			"keep_ratio":True,
			"source":url
		})
		w = blocks.widgetBuild(desc)
		if w is None:
			print('Error desc=',desc)
			return
		bl = BoxLayout(size_hint=(None,None),size=(self.box_width,self.box_width))
		bl.add_widget(w)
		self.add_widget(bl,**kw)
		print('showObject():add widget',desc,w.size)
		return bl





from ..baseWidget import Text, SelectInput
from ..utils import *

from .factory import UiFactory, get_value

def build_input_code_widget(desc, rec=None):
	v = get_value(desc, rec=rec)
	kw = desc.get('uiparams', {})
	w = SelectInput(**kw)
	if v:
		w.setValue(v)
	return w
	
def build_view_code_widget(desc, rec=None):
		tf = desc.get('textField','text')
		vf = desc.get('valueField','value')
		v = rec.get(tf,rec.get(vf, ' '))
		return Text(text=v, 
					i18n=True,
					font_size = CSize(1), 
					wrap=True,
					halign='center', 
					valign='middle'
		)

UiFactory.register('code', build_input_code_widget, build_view_code_widget)




import os
import kivy
import kivyblocks
from kivy.config import Config
from kivy.resources import resource_add_path
resource_add_path(os.path.join(os.path.dirname(kivyblocks.__file__),'./ttf'))
Config.set('kivy', 'default_font', [
        'msgothic',
        'DroidSansFallback.ttf'])

config_set = True




from kivy.uix.label import Label
from kivy.properties import NumericProperty, ColorProperty
from kivy.factory import Factory

from .baseWidget import FloatInput

from .uitype.factory import UiFactory, get_value

class PriceView(Label):
	dec = NumericProperty(4)
	compare_price = NumericProperty(None)
	price = NumericProperty(None)
	nor_color = ColorProperty([0.9,0.9,0.9,1]) # white
	inc_color = ColorProperty([0.9,0,0,1]) 	# red
	dec_color = ColorProperty([0,0.9,0,1])	# green
	def __init__(self, **kw):
		super(PriceView, self).__init__(text='', **kw)

	def on_compare_price(self, *args):
		self.on_price()
	def on_price(self, *args):
		if self.price is None:
			return
		if not isinstance(self.price, float) and \
					not isinstance(self.price, int):
			return

		f = '%%.0%df' % self.dec
		self.text = f % self.price
		if self.compare_price is None:
			self.color = self.nor_color
		elif self.price > self.compare_price:
			self.color = self.inc_color
		elif self.price < self.compare_price:
			self.color = self.dec_color
		else:
			self.color = nor_color

	def getValue(self):
		return self.price

	def setValue(self, v):
		self.price = v

	@classmethod
	def builder(desc, rec=None):
		v = get_value(desc, rec=rec)
		return PriceView(price=v,
						font_size=CSize(1),
						halign='right',
						valign='middle'
		)

class PriceInput(FloatInput):
	@classmethod
	def builder(self, desc, rec=None):
		v = get_value(desc, rec=rec)
		kw = desc.get('uiparams', {})
		w = PriceInput(**kw)
		if v is not None:
			w.setValue(v)
		return w
		
Factory.register('PriceView', PriceView)


UiFactory.register('price', PriceInput.builder, PriceView.builder)



import os
from appPublic.folderUtils import listFile

from kivy.utils import platform
from .paging import PageLoader

class FileLoader(PageLoader):
	def __init__(self, folder='/DCIM',**options):
		self.files = []
		fold = os.path.abspath(folder)
		self.params = options.get('params',{})
		suffixes = self.parasm.get('suffixes',[])
		rescursive = self.params.get('rescursive',True)
		suffixs=[],rescursive=False
		for f in listFile(fold, suffixs=suffixes, rescursive=resursive):
			x = {
				"id":f,
				"name":f
			}
			self.files.append(x)
		self.page_rows = options.get('page_rows',1)
		self.loading = False
		self.total_cnt = len(self.files)
		self.total_page = self.total_cnt / self.page_rows 
		if self.total_cnt % self.page_rows != 0:
			self.total_page += 1
		self.curpage = 0

	def loadPage(self,p):
		if p < 1 or p > self.total_page:
			return None
		beg = self.page_rows * (p - 1)
		end = beg + self.page_rows
		return self.files[beg:end]





# coding=utf-8
"""
Geojson layer
=============

.. note::

    Currently experimental and a work in progress, not fully optimized.


Supports:

- html color in properties
- polygon geometry are cached and not redrawed when the parent mapview changes
- linestring are redrawed everymove, it's ugly and slow.
- marker are NOT supported

"""

__all__ = ["GeoJsonMapLayer"]

import json

from kivy.graphics import (
    Canvas,
    Color,
    Line,
    MatrixInstruction,
    Mesh,
    PopMatrix,
    PushMatrix,
    Scale,
    Translate,
)
from kivy.graphics.tesselator import TYPE_POLYGONS, WINDING_ODD, Tesselator
from kivy.metrics import dp
from kivy.properties import ObjectProperty, StringProperty
from kivy.utils import get_color_from_hex

from kivyblocks.mapview.constants import CACHE_DIR
from kivyblocks.mapview.downloader import Downloader
from kivyblocks.mapview.view import MapLayer

COLORS = {
    'aliceblue': '#f0f8ff',
    'antiquewhite': '#faebd7',
    'aqua': '#00ffff',
    'aquamarine': '#7fffd4',
    'azure': '#f0ffff',
    'beige': '#f5f5dc',
    'bisque': '#ffe4c4',
    'black': '#000000',
    'blanchedalmond': '#ffebcd',
    'blue': '#0000ff',
    'blueviolet': '#8a2be2',
    'brown': '#a52a2a',
    'burlywood': '#deb887',
    'cadetblue': '#5f9ea0',
    'chartreuse': '#7fff00',
    'chocolate': '#d2691e',
    'coral': '#ff7f50',
    'cornflowerblue': '#6495ed',
    'cornsilk': '#fff8dc',
    'crimson': '#dc143c',
    'cyan': '#00ffff',
    'darkblue': '#00008b',
    'darkcyan': '#008b8b',
    'darkgoldenrod': '#b8860b',
    'darkgray': '#a9a9a9',
    'darkgrey': '#a9a9a9',
    'darkgreen': '#006400',
    'darkkhaki': '#bdb76b',
    'darkmagenta': '#8b008b',
    'darkolivegreen': '#556b2f',
    'darkorange': '#ff8c00',
    'darkorchid': '#9932cc',
    'darkred': '#8b0000',
    'darksalmon': '#e9967a',
    'darkseagreen': '#8fbc8f',
    'darkslateblue': '#483d8b',
    'darkslategray': '#2f4f4f',
    'darkslategrey': '#2f4f4f',
    'darkturquoise': '#00ced1',
    'darkviolet': '#9400d3',
    'deeppink': '#ff1493',
    'deepskyblue': '#00bfff',
    'dimgray': '#696969',
    'dimgrey': '#696969',
    'dodgerblue': '#1e90ff',
    'firebrick': '#b22222',
    'floralwhite': '#fffaf0',
    'forestgreen': '#228b22',
    'fuchsia': '#ff00ff',
    'gainsboro': '#dcdcdc',
    'ghostwhite': '#f8f8ff',
    'gold': '#ffd700',
    'goldenrod': '#daa520',
    'gray': '#808080',
    'grey': '#808080',
    'green': '#008000',
    'greenyellow': '#adff2f',
    'honeydew': '#f0fff0',
    'hotpink': '#ff69b4',
    'indianred': '#cd5c5c',
    'indigo': '#4b0082',
    'ivory': '#fffff0',
    'khaki': '#f0e68c',
    'lavender': '#e6e6fa',
    'lavenderblush': '#fff0f5',
    'lawngreen': '#7cfc00',
    'lemonchiffon': '#fffacd',
    'lightblue': '#add8e6',
    'lightcoral': '#f08080',
    'lightcyan': '#e0ffff',
    'lightgoldenrodyellow': '#fafad2',
    'lightgray': '#d3d3d3',
    'lightgrey': '#d3d3d3',
    'lightgreen': '#90ee90',
    'lightpink': '#ffb6c1',
    'lightsalmon': '#ffa07a',
    'lightseagreen': '#20b2aa',
    'lightskyblue': '#87cefa',
    'lightslategray': '#778899',
    'lightslategrey': '#778899',
    'lightsteelblue': '#b0c4de',
    'lightyellow': '#ffffe0',
    'lime': '#00ff00',
    'limegreen': '#32cd32',
    'linen': '#faf0e6',
    'magenta': '#ff00ff',
    'maroon': '#800000',
    'mediumaquamarine': '#66cdaa',
    'mediumblue': '#0000cd',
    'mediumorchid': '#ba55d3',
    'mediumpurple': '#9370d8',
    'mediumseagreen': '#3cb371',
    'mediumslateblue': '#7b68ee',
    'mediumspringgreen': '#00fa9a',
    'mediumturquoise': '#48d1cc',
    'mediumvioletred': '#c71585',
    'midnightblue': '#191970',
    'mintcream': '#f5fffa',
    'mistyrose': '#ffe4e1',
    'moccasin': '#ffe4b5',
    'navajowhite': '#ffdead',
    'navy': '#000080',
    'oldlace': '#fdf5e6',
    'olive': '#808000',
    'olivedrab': '#6b8e23',
    'orange': '#ffa500',
    'orangered': '#ff4500',
    'orchid': '#da70d6',
    'palegoldenrod': '#eee8aa',
    'palegreen': '#98fb98',
    'paleturquoise': '#afeeee',
    'palevioletred': '#d87093',
    'papayawhip': '#ffefd5',
    'peachpuff': '#ffdab9',
    'peru': '#cd853f',
    'pink': '#ffc0cb',
    'plum': '#dda0dd',
    'powderblue': '#b0e0e6',
    'purple': '#800080',
    'red': '#ff0000',
    'rosybrown': '#bc8f8f',
    'royalblue': '#4169e1',
    'saddlebrown': '#8b4513',
    'salmon': '#fa8072',
    'sandybrown': '#f4a460',
    'seagreen': '#2e8b57',
    'seashell': '#fff5ee',
    'sienna': '#a0522d',
    'silver': '#c0c0c0',
    'skyblue': '#87ceeb',
    'slateblue': '#6a5acd',
    'slategray': '#708090',
    'slategrey': '#708090',
    'snow': '#fffafa',
    'springgreen': '#00ff7f',
    'steelblue': '#4682b4',
    'tan': '#d2b48c',
    'teal': '#008080',
    'thistle': '#d8bfd8',
    'tomato': '#ff6347',
    'turquoise': '#40e0d0',
    'violet': '#ee82ee',
    'wheat': '#f5deb3',
    'white': '#ffffff',
    'whitesmoke': '#f5f5f5',
    'yellow': '#ffff00',
    'yellowgreen': '#9acd32',
}


def flatten(lst):
    return [item for sublist in lst for item in sublist]


class GeoJsonMapLayer(MapLayer):

    source = StringProperty()
    geojson = ObjectProperty()
    cache_dir = StringProperty(CACHE_DIR)

    def __init__(self, **kwargs):
        self.first_time = True
        self.initial_zoom = None
        super().__init__(**kwargs)
        with self.canvas:
            self.canvas_polygon = Canvas()
            self.canvas_line = Canvas()
        with self.canvas_polygon.before:
            PushMatrix()
            self.g_matrix = MatrixInstruction()
            self.g_scale = Scale()
            self.g_translate = Translate()
        with self.canvas_polygon:
            self.g_canvas_polygon = Canvas()
        with self.canvas_polygon.after:
            PopMatrix()

    def reposition(self):
        vx, vy = self.parent.delta_x, self.parent.delta_y
        pzoom = self.parent.zoom
        zoom = self.initial_zoom
        if zoom is None:
            self.initial_zoom = zoom = pzoom
        if zoom != pzoom:
            diff = 2 ** (pzoom - zoom)
            vx /= diff
            vy /= diff
            self.g_scale.x = self.g_scale.y = diff
        else:
            self.g_scale.x = self.g_scale.y = 1.0
        self.g_translate.xy = vx, vy
        self.g_matrix.matrix = self.parent._scatter.transform

        if self.geojson:
            update = not self.first_time
            self.on_geojson(self, self.geojson, update=update)
            self.first_time = False

    def traverse_feature(self, func, part=None):
        """Traverse the whole geojson and call the func with every element
        found.
        """
        if part is None:
            part = self.geojson
        if not part:
            return
        tp = part["type"]
        if tp == "FeatureCollection":
            for feature in part["features"]:
                func(feature)
        elif tp == "Feature":
            func(part)

    @property
    def bounds(self):
        # return the min lon, max lon, min lat, max lat
        bounds = [float("inf"), float("-inf"), float("inf"), float("-inf")]

        def _submit_coordinate(coord):
            lon, lat = coord
            bounds[0] = min(bounds[0], lon)
            bounds[1] = max(bounds[1], lon)
            bounds[2] = min(bounds[2], lat)
            bounds[3] = max(bounds[3], lat)

        def _get_bounds(feature):
            geometry = feature["geometry"]
            tp = geometry["type"]
            if tp == "Point":
                _submit_coordinate(geometry["coordinates"])
            elif tp == "Polygon":
                for coordinate in geometry["coordinates"][0]:
                    _submit_coordinate(coordinate)
            elif tp == "MultiPolygon":
                for polygon in geometry["coordinates"]:
                    for coordinate in polygon[0]:
                        _submit_coordinate(coordinate)

        self.traverse_feature(_get_bounds)
        return bounds

    @property
    def center(self):
        min_lon, max_lon, min_lat, max_lat = self.bounds
        cx = (max_lon - min_lon) / 2.0
        cy = (max_lat - min_lat) / 2.0
        return min_lon + cx, min_lat + cy

    def on_geojson(self, instance, geojson, update=False):
        if self.parent is None:
            return
        if not update:
            self.g_canvas_polygon.clear()
            self._geojson_part(geojson, geotype="Polygon")
        self.canvas_line.clear()
        self._geojson_part(geojson, geotype="LineString")

    def on_source(self, instance, value):
        if value.startswith(("http://", "https://")):
            Downloader.instance(cache_dir=self.cache_dir).download(
                value, self._load_geojson_url
            )
        else:
            with open(value, "rb") as fd:
                geojson = json.load(fd)
            self.geojson = geojson

    def _load_geojson_url(self, url, response):
        self.geojson = response.json()

    def _geojson_part(self, part, geotype=None):
        tp = part["type"]
        if tp == "FeatureCollection":
            for feature in part["features"]:
                if geotype and feature["geometry"]["type"] != geotype:
                    continue
                self._geojson_part_f(feature)
        elif tp == "Feature":
            if geotype and part["geometry"]["type"] == geotype:
                self._geojson_part_f(part)
        else:
            # unhandled geojson part
            pass

    def _geojson_part_f(self, feature):
        properties = feature["properties"]
        geometry = feature["geometry"]
        graphics = self._geojson_part_geometry(geometry, properties)
        for g in graphics:
            tp = geometry["type"]
            if tp == "Polygon":
                self.g_canvas_polygon.add(g)
            else:
                self.canvas_line.add(g)

    def _geojson_part_geometry(self, geometry, properties):
        tp = geometry["type"]
        graphics = []
        if tp == "Polygon":
            tess = Tesselator()
            for c in geometry["coordinates"]:
                xy = list(self._lonlat_to_xy(c))
                xy = flatten(xy)
                tess.add_contour(xy)

            tess.tesselate(WINDING_ODD, TYPE_POLYGONS)

            color = self._get_color_from(properties.get("color", "FF000088"))
            graphics.append(Color(*color))
            for vertices, indices in tess.meshes:
                graphics.append(
                    Mesh(vertices=vertices, indices=indices, mode="triangle_fan")
                )

        elif tp == "LineString":
            stroke = get_color_from_hex(properties.get("stroke", "#ffffff"))
            stroke_width = dp(properties.get("stroke-width"))
            xy = list(self._lonlat_to_xy(geometry["coordinates"]))
            xy = flatten(xy)
            graphics.append(Color(*stroke))
            graphics.append(Line(points=xy, width=stroke_width))

        return graphics

    def _lonlat_to_xy(self, lonlats):
        view = self.parent
        zoom = view.zoom
        for lon, lat in lonlats:
            p = view.get_window_xy_from(lat, lon, zoom)
            p = p[0] - self.parent.delta_x, p[1] - self.parent.delta_y
            p = self.parent._scatter.to_local(*p)
            yield p

    def _get_color_from(self, value):
        color = COLORS.get(value.lower(), value)
        color = get_color_from_hex(color)
        return color



# coding=utf-8

__all__ = ["clamp", "haversine", "get_zoom_for_radius"]

from math import asin, cos, pi, radians, sin, sqrt

from kivy.core.window import Window
from kivy.metrics import dp


def clamp(x, minimum, maximum):
    return max(minimum, min(x, maximum))


def haversine(lon1, lat1, lon2, lat2):
    """
    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees)

    Taken from: http://stackoverflow.com/questions/4913349/haversine-formula-in-python-bearing-and-distance-between-two-gps-points
    """
    # convert decimal degrees to radians
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])
    # haversine formula
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat / 2) ** 2 + cos(lat1) * cos(lat2) * sin(dlon / 2) ** 2

    c = 2 * asin(sqrt(a))
    km = 6367 * c
    return km


def get_zoom_for_radius(radius_km, lat=None, tile_size=256.0):
    """See: https://wiki.openstreetmap.org/wiki/Zoom_levels"""
    radius = radius_km * 1000.0
    if lat is None:
        lat = 0.0  # Do not compensate for the latitude

    # Calculate the equatorial circumference based on the WGS-84 radius
    earth_circumference = 2.0 * pi * 6378137.0 * cos(lat * pi / 180.0)

    # Check how many tiles that are currently in view
    nr_tiles_shown = min(Window.size) / dp(tile_size)

    # Keep zooming in until we find a zoom level where the circle can fit inside the screen
    zoom = 1
    while earth_circumference / (2 << (zoom - 1)) * nr_tiles_shown > 2 * radius:
        zoom += 1
    return zoom - 1  # Go one zoom level back




from ..baseWidget import WrapText, StrInput
from ..utils import *
from .factory import UiFactory, get_value

class TextView(WrapText):
	def getValue(self):
		return self.text

	def setValue(self, v):
		self.text = v

	@classmethod
	def builder(self, desc, rec=None):
		txt = get_value(desc, rec=rec)
		return TextView(text=str(txt) if txt else '',
						font_size=CSize(1),
						halign='left',
						valign='middle')

def build_input_text_widget(desc, rec=None):
	v = get_value(desc, rec=rec)
	kw = desc.get('uiparams', {})
	w = StrInput(**kw)
	if v:
		self.multiline = True
		w.setValue(v)
	return w
	
UiFactory.register('text', build_input_text_widget, TextView.builder)




from kivy.factory import Factory
from kivy.uix.widget import Widget
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.clock import Clock
from kivyblocks.swipebehavior import SwipeBehavior
from kivyblocks.ready import WidgetReady

class OrientationLayout(FloatLayout, WidgetReady, SwipeBehavior):
	"""
	TwinStyleLayout layout two widget verical when parital orientation
	and second widget hide when it is in landscape
	"""
	def __init__(self,	main_widget=None, second_widget=None, **kw):
		self.main_widget = main_widget
		self.second_widget = second_widget
		self.widget_main = None
		self.widget_second = None
		self.second_flg = False
		super(OrientationLayout, self).__init__(**kw)
		self.build_children()
		self.bind(on_swipe_left=self.toggle_second)
		self.bind(on_swipe_right=self.toggle_second)
		self.bind(size=self.on_size_changed)
		self.bind(pos=self.on_size_changed)
		self.second_showed = None
		self.reready()
		self.register_event_type('on_interactive')
	
	def build_children(self, *args):
		blocks = Factory.Blocks()
		self.widget_main = blocks.widgetBuild(self.main_widget)
		blocks = Factory.Blocks()
		self.widget_second = blocks.widgetBuild(self.second_widget)

	def isLandscape(self):
		return self.width > self.height

	def toggle_second(self,*args):
		if self.isLandscape():
			if self.second_flg:
				self.remove_widget(self.widget_second)
				self.second_flg = False
				self.dispatch('on_interactive')
			else:
				self.add_widget(self.widget_second)
				self.second_flg = True
				self.on_size_changed(self.size)
				self.dispatch('on_interactive')

	def on_size_changed(self,*args):
		Clock.schedule_once(self.two_widget_layout,0)

	def on_interactive(self, *args):
		print('on_orientation_changed fired')

	def two_widget_layout(self, *args):
		if not isinstance(self.widget_main, Widget) or \
				not isinstance(self.widget_second, Widget):
			# Clock.schedule_once(self.two_widget_layout,0)
			return

		if self.isLandscape():
			self.horizontal_layout()
		else:
			self.vertical_layout()
		self.do_layout()

	def horizontal_layout(self):
		self.widget_main.size_hint = (None,None)
		self.widget_main.size = self.size
		self.widget_main.pos = self.pos
		self.widget_second.height = self.height
		self.widget_second.size_hint = (None,None)
		self.widget_second.width = self.width * self.height / self.width
		self.widget_second.pos = (0,0)
		self.widget_second.opacity = 0.6
		self.clear_widgets()
		self.add_widget(self.widget_main)
		if self.second_flg:
			self.add_widget(self.widget_second)

	def vertical_layout(self):
		self.widget_main.size_hint = (None,None)
		self.widget_main.size = (self.width, \
										self.width / 16 * 10)
		self.widget_main.pos = (0, self.height - self.widget_main.height)
		self.widget_second.size_hint = (None, None)
		self.widget_second.size = (self.width, \
							self.height - self.widget_main.height)
		self.widget_second.pos = (0,0)
		self.widget_second.opacity = 0.6
		self.clear_widgets()
		self.add_widget(self.widget_main)
		self.add_widget(self.widget_second)

	def widget_build_failed(self, o, e):
		pass




from jnius import autoclass

def startService(className):
	# service = autoclass('your.service.name.ClassName')
	service = autoclass(className)
	mActivity = autoclass('org.kivy.android.PythonActivity').mActivity
	service.start(mActivity)

def stopService(className):
	service = autoclass(className)
	mActivity = autoclass('org.kivy.android.PythonActivity').mActivity
	service.stop(mActivity)



from kivy.app import App
from kivy.metrics import mm
from kivy.core.window import Window

from appPublic.Singleton import SingletonDecorator
from appPublic.jsonConfig import getConfig

@SingletonDecorator
class KivySizes:
	myFontSizes = {
		"smallest":1.5,
		"small":2.0,
		"normal":2.5,
		"large":3.0,
		"huge":3.5,
		"hugest":4.0,
	}
	separatorsize = 2

	def getFontSize(self,name=None):
		config = getConfig()
		if config.font_sizes:
			self.myFontSizes = config.font_sizes
		if name is None:
			name = config.font_name
		x = self.myFontSizes.get(name,None)
		if x == None:
			x = self.myFontSizes.get('normal')
		return x

	def namedSize(self,cnt=1,name=None):
		return mm(cnt * self.getFontSize(name=name))

	def unitedSize(self,x,y=None,name=None):
		xr = self.namedSize(cnt=x,name=name)
		if y is None:
			return xr
		return (xr,self.namedSize(cnt=y,name=name))

	def CSize(self,x,y=None,name=None):
		return self.unitedSize(x,y=y,name=name)

	def getScreenSize(self):
		return Window.width, Window.height

	def getWindowPhysicalSize(self):
		h_phy = float(Window.height) / mm(1)
		w_phy = float(Window.width) / mm(1)
		return w_phy, h_phy
	
	def getScreenPhysicalSize(self):
		return self.getWindowPhysicalSize()

	def isHandHold(self):
		return min(self.getScreenPhysicalSize()) <= 75.0




import requests

url = "http://localhost:8080/uploadfile.dspy"

files = {
	"afile":("uploadfile.py",open("uploadfile.py","rb"),"application/python")
}

r = requests.post(url,files=files)
print(r.text)



import kivy
from kivy.utils import platform
from kivy.uix.textinput import TextInput

from appPublic.registerfunction import RegisterFunction

import kivyblocks.clickable
import kivyblocks.multi_select

from .baseWidget import *
from .tree import Tree, TextTree, MenuTree, PopupMenu
from .toolbar import ToolPage, Toolbar
from .dg import DataGrid
from .vplayer import VPlayer
from .aplayer import APlayer
from .form import Form, StrSearchForm
from .boxViewer import BoxViewer
from .pagescontainer import PageContainer
from .hostimage import HostImage
from .toggleitems import PressableBox, ToggleItems
from .twosides import TwoSides
from .tab import TabsPanel
from .qrdata import QRCodeWidget
# from .kivycamera import KivyCamera
from .filebrowser import FileLoaderBrowser
from .mapview import MapView
from .message import Conform
from .pagepanel import PagePanel
from .markdown import Markdown
# from .custom_camera import CustomCamera, QrReader
from .defaultimage import *
from .price import *

#if kivy.platform in ['win','linux', 'macosx']:
#	from .camerawithmic import ScreenWithMic
#from .camerawithmic import CameraWithMic
from .scrollpanel import ScrollPanel
from .udp_widget import UdpWidget
from .paging import PageLoader
from .dateinput import DateInput
from .block_test import BlockTest
from .hierarchy import Hierarchy
from .price import PriceView
from .ffpyplayer_video import FFVideo

r = Factory.register
# if kivy.platform in ['win','linux', 'macosx']:
#	r('ScreenWithMic', ScreenWithMic)
r('FFVideo', FFVideo)
r('AnchorBox', AnchorBox)
r('FloatBox', FloatBox)
r('RelativeBox', RelativeBox)
r('GridBox', GridBox)
r('PageBox', PageBox)
r('ScatterBox', ScatterBox)
r('StackBox', StackBox)
r('DateInput', DateInput)
r('HTTPSeriesData', HTTPSeriesData)
r('HTTPDataHandler', HTTPDataHandler)
r('PageLoader', PageLoader)
r('UdpWidget', UdpWidget)
r('ScrollPanel', ScrollPanel)
r('TextInput', TextInput)
# r('CameraWithMic', CameraWithMic)
# r('CustomCamera', CustomCamera)
# r('QrReader', QrReader)
r('Markdown', Markdown)
r('PagePanel', PagePanel)
r('Conform', Conform)
r('Popup', Popup)
r('MapView', MapView)
r('DataGrid',DataGrid)
r('FileLoaderBrowser',FileLoaderBrowser)
# r('KivyCamera',KivyCamera)
r('QRCodeWidget',QRCodeWidget)
r('TabsPanel',TabsPanel)
r('TwoSides',TwoSides)
r('PageContainer', PageContainer)
r('BoxViewer', BoxViewer)
r('Form', Form)
r('StrSearchForm', StrSearchForm)
r('VPlayer', VPlayer)
r('DataGrid', DataGrid)
r('Toolbar', Toolbar)
r('ToolPage',ToolPage)
r('HTTPDataHandler',HTTPDataHandler)
r('Text',Text)
r('ScrollWidget',ScrollWidget)
r('BinStateImage',BinStateImage)
r('JsonCodeInput',JsonCodeInput)
r('FloatInput',FloatInput)
r('IntegerInput',IntegerInput)
r('StrInput',StrInput)
r('SelectInput',SelectInput)
r('BoolInput',BoolInput)
r('Messager',Messager)
r('LoginForm',LoginForm)
r('PressableImage', PressableImage)
r('PressableLabel', PressableLabel)
r('Tree',Tree)
r('TextTree',TextTree)
r('MenuTree',MenuTree)
r('PopupMenu',PopupMenu)
r('HostImage',HostImage)
r('APlayer',APlayer)
r('WrapText',WrapText)
r('PressableBox',PressableBox)
r('Title1',Title1)
r('Title2',Title2)
r('Title3',Title3)
r('Title4',Title4)
r('Title5',Title5)
r('Title6',Title6)
r('Modal',Modal)
r('TimedModal',TimedModal)
r('HBox',HBox)
r('VBox',VBox)
r('SwipeBox',SwipeBox)
r('ToggleItems',ToggleItems)
r('ExAccordion', ExAccordion)
r('Slider', Slider)
if platform == 'android':
	r('PhoneButton',PhoneButton)
	r('AWebView',AWebView)


def register_widget(name, klass):
	try:
		Factory.regiter(name, klass)
	except:
		Logger.info(f'Plugin : register_widget():{name} register error')

def register_registerfunction(name, func):
	rf = RegisterFunction()
	try:
		rf.register(name, func)
	except Exception as e:
		Logger.info(f'Plugin : register_registerfunction():{name} register error({e})')
		print_exc()

def register_blocks(name, value):
	b = Factory.Blocks()
	try:
		b.register_widget(name, value)
	except:
		Logger.info(f'plugin : register_blocks():{name} register error')






from math import floor
from kivy.properties import NumericProperty
from kivy.utils import platform
from kivy.uix.scrollview import ScrollView
from kivy.uix.gridlayout import GridLayout
from kivy.graphics import Color, Ellipse,Rectangle
from kivy.clock import Clock
from .utils import CSize, isHandHold, show_widget_info
from .widget_css import WidgetCSS
from .ready import WidgetReady

class VResponsiveLayout(WidgetCSS, WidgetReady, ScrollView):
	box_width = NumericProperty(None)
	box_width_c = NumericProperty(None)
	columes = NumericProperty(None)
	def __init__(self, **kw):
		self._inner = None
		self.col_width = None
		self.cols = None
		super(VResponsiveLayout, self).__init__(**kw)
		self.options = kw
		self._inner = GridLayout(cols=1, padding=2, 
						spacing=2,size_hint=(1,None))
		self._inner.col_force_default = True
		self._inner.col_default_width = self.box_width
		super(VResponsiveLayout,self).add_widget(self._inner)
		self._inner.bind(
				minimum_height=self._inner.setter('height'))
		self.bind(size=self.set_col_width)
	
	def calculate_col_width(self, cnt):
		# cnt * col_width + 2*padding + (cnt-1) * spacing = width
		w = self._inner
		padding = 0
		if isinstance(w.padding, float):
			padding = w.padding
		elif len(w.padding) < 4:
			padding = w.padding[0]
		else:
			padding = (w.padding[0] + w.padding[2])/2
		spacing = 0
		if isinstance(w.spacing, float):
			spacing = w.spacing
		elif len(w.spacing) < 4:
			spacing = w.spacing[0]
		else:
			spacing = (w.spacing[0] + w.spacing[2]) / 2
		width = (w.width - 2 * padding - cnt * spacing) / cnt
		return width

	def get_col_width(self):
		return self._inner.col_default_width

	def get_cols(self):
		return self._inner.cols

	def set_col_width(self, o, s):
		if self._inner is None:
			return
		self._inner.size_hint = [None, None]
		self._inner.size = self.size
		if isHandHold() and self.width < self.height:
			self.cols = 1
			self.col_width = self.calculate_col_width(self.cols)
			return self.setCols()

		if self.columes is not None and self.columes > 0:
			self.cols = self.columes
			self.col_width = self.calculate_col_width(self.cols)
			return self.setCols()
			
		if self.box_width_c is not None:
			self.cols = floor(self._inner.width / CSize(self.box_width_c))
			if self.cols < 1:
				self.cols = 1
			self.col_width = self.calculate_col_width(self.cols)
			return self.setCols()

		if self.box_width is not None:
			if self.box_width <= 1:
				self.cols = floor(1 / self.box_width)
				self.col_width = self.calculate_col_width(self.cols)
				return self.setCols()
			else:
				self.cols = floor(self._inner.width / self.box_width)
				if self.cols < 1:
					self.cols = 1
				self.col_width = self.calculate_col_width(self.cols)
				return self.setCols()
		return
			
	def on_orientation(self,o):
		self.set_col_width()

	def add_widget(self,widget,**kw):
		a = self._inner.add_widget(widget,**kw)
		return a

	def clear_widgets(self,**kw):
		a = self._inner.clear_widgets(**kw)

	def remove_widget(self,widget,**kw):
		a = self._inner.remove_widget(widget,**kw)
		return a

	def setCols(self,*args):
		self._inner.cols = self.cols
		self._inner.col_default_width = self.col_width
		for w in self._inner.children:
			w.size_hint_x = None
			w.width = self._inner.col_default_width




# -*- coding=utf-8 -*-

import time
from threading import Thread, Lock, BoundedSemaphore
import requests
from functools import wraps

from kivy.event import EventDispatcher
from kivy.clock import Clock
from kivy.app import App
from .login import LoginForm

from appPublic.http_client import Http_Client

class ThreadCall(Thread,EventDispatcher):
	def __init__(self,target, args=(), kwargs={}):
		Thread.__init__(self)
		EventDispatcher.__init__(self)
		self.register_event_type('on_result')
		self.register_event_type('on_error')
		self.rez = None
		self.daemon = False
		self.target = target
		self.args = args
		self.timing = None
		self.kwargs = kwargs
	
	def start(self):
		Thread.start(self)
		self.timing = Clock.schedule_once(self.checkStop,0)

	def run(self):
		try:
			# print('ThreadCall()',self.args,'start...')
			self.rez = self.target(*self.args,**self.kwargs)
			self.dispatch('on_result',self.rez)
			# print('ThreadCall()',*self.args,'finished...')

		except Exception as e:
			# print('ThreadCall()',*self.args,'Error...')
			self.dispatch('on_error',e)

	def on_result(self, v):
		pass # print('ThreadCall():on_result() called,v=',v)

	def on_error(self,e):
		pass

	def checkStop(self,timestamp):
		x = self.join(timeout=0.001)
		if self.is_alive():
			self.timing = Clock.schedule_once(self.checkStop,0)
			return

class Workers(Thread):
	def __init__(self,maxworkers):
		super().__init__()
		self.max_workers = maxworkers
		self.tasks = []
		# task = [callee,callback,kwargs]
		self.lock = Lock()
		self.work_sema = BoundedSemaphore(value=self.max_workers)
		self.running = False
	def run(self):
		self.running = True
		while self.running:
			if len(self.tasks) == 0:
				time.sleep(0.001)
				continue
			task = None
			with self.lock:
				task = self.tasks.pop()
			if task is None:
				continue
			with self.work_sema:
				callee,callback,errback,kwargs = task
				x = ThreadCall(callee,kwargs=kwargs)
				x.bind(on_result=callback)
				"""
				并发的时候，只有一个callback会被调用
				"""
				if errback:
					x.bind(on_error=errback)
				x.start()

	def add(self,callee,callback,errback=None,kwargs={}):
		with self.lock:
			self.tasks.insert(0,[callee,callback,errback,kwargs])

class HttpClient(Http_Client):
	def __init__(self):
		super().__init__()
		self.workers = App.get_running_app().workers
		
	def __call__(self,url,method="GET",
				params={},
				headers={},
				files={},
				stream=False,
				callback=None,
				errback=None):
		method = method.upper()
		def cb(t,resp):
			return resp

		if callback is None:
			try:
				resp = self.webcall(url, method=method,
						params=params, files=files, headers=headers)
				return cb(None,resp)
			except Exception as e:
				raise e
		kwargs = {
			"url":url,
			"method":method,
			"params":params,
			"files":files,
			"stream":stream,
			"headers":headers
		}

		self.workers.add(self.webcall,callback,errback,kwargs=kwargs)

	def get(self, url, params={}, headers={}, stream=False, callback=None, errback=None):
		return self.__call__(url,method='GET',params=params,
				headers=headers, stream=stream, callback=callback,
				errback=errback)
	def post(self, url, params={}, headers={}, files={}, stream=False, callback=None, errback=None):
		return self.__call__(url,method='POST',params=params, files=files,
				headers=headers, stream=stream, callback=callback,
				errback=errback)

	def put(self, url, params={}, headers={}, stream=False, callback=None, errback=None):
		return self.__call__(url,method='PUT',params=params,
				headers=headers, stream=stream, callback=callback,
				errback=errback)

	def delete(self, url, params={}, headers={}, callback=None, errback=None):
		return self.__call__(url,method='DELETE',params=params,
				headers=headers, callback=callback,
				errback=errback)

	def option(self, url, params={}, headers={}, callback=None, errback=None):
		return self.__call__(url,method='OPTION',params=params,
				headers=headers, callback=callback, errback=errback)
	
if __name__ == '__main__':
	from kivy.uix.textinput import TextInput
	from kivy.app import App
	from kivy.uix.boxlayout import BoxLayout
	from kivy.uix.button import Button
	class MyApp(App):
		def build(self):
			self.hc = HttpClient()
			x = BoxLayout(orientation='vertical')
			y = BoxLayout(orientation='horizontal',size_hint_y=0.07)
			self.ti = TextInput(size_hint_x=0.95,multiline=False)
			btn = Button(size_hint_x=0.05,text='go')
			y.add_widget(self.ti)
			y.add_widget(btn)
			btn.bind(on_press=self.getHtml)
			self.texti = TextInput(multiline=True,readonly=True)
			x.add_widget(y)
			x.add_widget(self.texti)
			
			return x
	
		def getHtml(self,v=None):
			url = self.ti.text
			self.hc.get(url,callback=self.showResult)
			self.texti.text = 'loading...'

		def showResult(self,target,resp):
			if resp.status_code==200:
				self.texti.text = resp.text
			else:
				print(reps.status_code,'...............')
	MyApp().run()




from functools import partial
from kivy.logger import Logger
from kivy.factory import Factory

def get_cmd_desc(cmd_desc):
	desc = {
	}
	v = cmd_desc
	desc['conform'] = v.get('conform')
	desc['target'] = v.get('target')
	if v.get('datawidget'):
		desc['datawidget'] = v.get('datawidget')
		if v.get('datamethod'):
			desc['datamethod'] = v['datamethod']
	keys = v.keys()
	if 'url' in keys:
		Logger.info('get_cmd_desc():cmd_desc=%s', cmd_desc)
		desc['actiontype'] = 'urlwidget'
		desc['mode'] = 'replace'
		options = {
			'params':v.get('params',{}),
			'url':v.get('url')
		}
		desc['options'] = options
		return desc
	if 'rfname' in keys:
		desc['actiontype'] = 'registedfunction'
		desc['params'] = v.get('params',{})
		return desc
	if 'script' in keys:
		desc['actiontype'] = 'script'
		desc['script'] = v.get('script')
		return desc
	if 'method' in keys:
		desc['actiontype'] = 'method'
		desc['method'] = v.get('method')
		return desc
	if 'actions' in keys:
		desc['actiontype'] = 'multiple'
		desc['actions'] = v.get('actions')
		return desc
	return None

def cmd_action(cmd_desc, widget):
	desc = get_cmd_desc(cmd_desc)
	if desc is None:
		Logger.error('CommandAction: desc is None cmd_desc=%s error', \
						cmd_desc)
		return
	blocks = Factory.Blocks()
	conform_desc = desc.get('conform')
	if conform_desc is None:
		Logger.info('cmd_action():desc=%s', desc)
		blocks.uniaction(widget, desc)
		return

	w = blocks.widgetBuild({
			"widgettype":"Conform",
			"options":conform_desc
	})
	w.bind(on_conform=partial(blocks.uniaction, widget, desc))
	w.open()
	Logger.info('cmd_action():desc=%s, conform and action', desc)





# coding=utf-8

__all__ = ["MapView", "MapMarker", "MapMarkerPopup", "MapLayer", "MarkerMapLayer"]

import webbrowser
from itertools import takewhile
from math import ceil
from os.path import dirname, join

from kivy.clock import Clock
from kivy.compat import string_types
from kivy.graphics import Canvas, Color, Rectangle
from kivy.graphics.transformation import Matrix
from kivy.lang import Builder
from kivy.metrics import dp
from kivy.properties import (
    AliasProperty,
    BooleanProperty,
    ListProperty,
    NumericProperty,
    ObjectProperty,
    StringProperty,
)
from kivy.uix.behaviors import ButtonBehavior
from kivy.uix.image import Image
from kivy.uix.label import Label
from kivy.uix.scatter import Scatter
from kivy.uix.widget import Widget

from kivyblocks.mapview import Bbox, Coordinate
from kivyblocks.mapview.constants import (
    CACHE_DIR,
    MAX_LATITUDE,
    MAX_LONGITUDE,
    MIN_LATITUDE,
    MIN_LONGITUDE,
)
from kivyblocks.mapview.source import MapSource
from kivyblocks.mapview.utils import clamp

Builder.load_string(
    """
<MapMarker>:
    size_hint: None, None
    source: root.source
    size: list(map(dp, self.texture_size))
    allow_stretch: True

<MapView>:
    canvas.before:
        StencilPush
        Rectangle:
            pos: self.pos
            size: self.size
        StencilUse
        Color:
            rgba: self.background_color
        Rectangle:
            pos: self.pos
            size: self.size
    canvas.after:
        StencilUnUse
        Rectangle:
            pos: self.pos
            size: self.size
        StencilPop

    ClickableLabel:
        text: root.map_source.attribution if hasattr(root.map_source, "attribution") else ""
        size_hint: None, None
        size: self.texture_size[0] + sp(8), self.texture_size[1] + sp(4)
        font_size: "10sp"
        right: [root.right, self.center][0]
        color: 0, 0, 0, 1
        markup: True
        canvas.before:
            Color:
                rgba: .8, .8, .8, .8
            Rectangle:
                pos: self.pos
                size: self.size


<MapViewScatter>:
    auto_bring_to_front: False
    do_rotation: False
    scale_min: 0.2
    scale_max: 3.

<MapMarkerPopup>:
    RelativeLayout:
        id: placeholder
        y: root.top
        center_x: root.center_x
        size: root.popup_size

"""
)


class ClickableLabel(Label):
    def on_ref_press(self, *args):
        webbrowser.open(str(args[0]), new=2)


class Tile(Rectangle):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cache_dir = kwargs.get('cache_dir', CACHE_DIR)

    @property
    def cache_fn(self):
        map_source = self.map_source
        fn = map_source.cache_fmt.format(
            image_ext=map_source.image_ext,
            cache_key=map_source.cache_key,
            **self.__dict__
        )
        return join(self.cache_dir, fn)

    def set_source(self, cache_fn):
        self.source = cache_fn
        self.state = "need-animation"


class MapMarker(ButtonBehavior, Image):
    """A marker on a map, that must be used on a :class:`MapMarker`
    """

    anchor_x = NumericProperty(0.5)
    """Anchor of the marker on the X axis. Defaults to 0.5, mean the anchor will
    be at the X center of the image.
    """

    anchor_y = NumericProperty(0)
    """Anchor of the marker on the Y axis. Defaults to 0, mean the anchor will
    be at the Y bottom of the image.
    """

    lat = NumericProperty(0)
    """Latitude of the marker
    """

    lon = NumericProperty(0)
    """Longitude of the marker
    """

    source = StringProperty(join(dirname(__file__), "icons", "marker.png"))
    """Source of the marker, defaults to our own marker.png
    """

    # (internal) reference to its layer
    _layer = None

    def detach(self):
        if self._layer:
            self._layer.remove_widget(self)
            self._layer = None


class MapMarkerPopup(MapMarker):
    is_open = BooleanProperty(False)
    placeholder = ObjectProperty(None)
    popup_size = ListProperty([100, 100])

    def add_widget(self, widget):
        if not self.placeholder:
            self.placeholder = widget
            if self.is_open:
                super().add_widget(self.placeholder)
        else:
            self.placeholder.add_widget(widget)

    def remove_widget(self, widget):
        if widget is not self.placeholder:
            self.placeholder.remove_widget(widget)
        else:
            super().remove_widget(widget)

    def on_is_open(self, *args):
        self.refresh_open_status()

    def on_release(self, *args):
        self.is_open = not self.is_open

    def refresh_open_status(self):
        if not self.is_open and self.placeholder.parent:
            super().remove_widget(self.placeholder)
        elif self.is_open and not self.placeholder.parent:
            super().add_widget(self.placeholder)


class MapLayer(Widget):
    """A map layer, that is repositionned everytime the :class:`MapView` is
    moved.
    """

    viewport_x = NumericProperty(0)
    viewport_y = NumericProperty(0)

    def reposition(self):
        """Function called when :class:`MapView` is moved. You must recalculate
        the position of your children.
        """
        pass

    def unload(self):
        """Called when the view want to completly unload the layer.
        """
        pass


class MarkerMapLayer(MapLayer):
    """A map layer for :class:`MapMarker`
    """

    order_marker_by_latitude = BooleanProperty(True)

    def __init__(self, **kwargs):
        self.markers = []
        super().__init__(**kwargs)

    def insert_marker(self, marker, **kwargs):
        if self.order_marker_by_latitude:
            before = list(
                takewhile(lambda i_m: i_m[1].lat < marker.lat, enumerate(self.children))
            )
            if before:
                kwargs['index'] = before[-1][0] + 1

        super().add_widget(marker, **kwargs)

    def add_widget(self, marker):
        marker._layer = self
        self.markers.append(marker)
        self.insert_marker(marker)

    def remove_widget(self, marker):
        marker._layer = None
        if marker in self.markers:
            self.markers.remove(marker)
        super().remove_widget(marker)

    def reposition(self):
        if not self.markers:
            return
        mapview = self.parent
        set_marker_position = self.set_marker_position
        bbox = None
        # reposition the markers depending the latitude
        markers = sorted(self.markers, key=lambda x: -x.lat)
        margin = max((max(marker.size) for marker in markers))
        bbox = mapview.get_bbox(margin)
        for marker in markers:
            if bbox.collide(marker.lat, marker.lon):
                set_marker_position(mapview, marker)
                if not marker.parent:
                    self.insert_marker(marker)
            else:
                super().remove_widget(marker)

    def set_marker_position(self, mapview, marker):
        x, y = mapview.get_window_xy_from(marker.lat, marker.lon, mapview.zoom)
        marker.x = int(x - marker.width * marker.anchor_x)
        marker.y = int(y - marker.height * marker.anchor_y)

    def unload(self):
        self.clear_widgets()
        del self.markers[:]


class MapViewScatter(Scatter):
    # internal
    def on_transform(self, *args):
        super().on_transform(*args)
        self.parent.on_transform(self.transform)

    def collide_point(self, x, y):
        return True


class MapView(Widget):
    """MapView is the widget that control the map displaying, navigation, and
    layers management.
    """

    lon = NumericProperty()
    """Longitude at the center of the widget
    """

    lat = NumericProperty()
    """Latitude at the center of the widget
    """

    zoom = NumericProperty(0)
    """Zoom of the widget. Must be between :meth:`MapSource.get_min_zoom` and
    :meth:`MapSource.get_max_zoom`. Default to 0.
    """

    map_source = ObjectProperty(MapSource())
    """Provider of the map, default to a empty :class:`MapSource`.
    """

    double_tap_zoom = BooleanProperty(False)
    """If True, this will activate the double-tap to zoom.
    """

    pause_on_action = BooleanProperty(True)
    """Pause any map loading / tiles loading when an action is done.
    This allow better performance on mobile, but can be safely deactivated on
    desktop.
    """

    snap_to_zoom = BooleanProperty(True)
    """When the user initiate a zoom, it will snap to the closest zoom for
    better graphics. The map can be blur if the map is scaled between 2 zoom.
    Default to True, even if it doesn't fully working yet.
    """

    animation_duration = NumericProperty(100)
    """Duration to animate Tiles alpha from 0 to 1 when it's ready to show.
    Default to 100 as 100ms. Use 0 to deactivate.
    """

    delta_x = NumericProperty(0)
    delta_y = NumericProperty(0)
    background_color = ListProperty([181 / 255.0, 208 / 255.0, 208 / 255.0, 1])
    cache_dir = StringProperty(CACHE_DIR)
    _zoom = NumericProperty(0)
    _pause = BooleanProperty(False)
    _scale = 1.0
    _disabled_count = 0

    __events__ = ["on_map_relocated"]

    # Public API

    @property
    def viewport_pos(self):
        vx, vy = self._scatter.to_local(self.x, self.y)
        return vx - self.delta_x, vy - self.delta_y

    @property
    def scale(self):
        if self._invalid_scale:
            self._invalid_scale = False
            self._scale = self._scatter.scale
        return self._scale

    def get_bbox(self, margin=0):
        """Returns the bounding box from the bottom/left (lat1, lon1) to
        top/right (lat2, lon2).
        """
        x1, y1 = self.to_local(0 - margin, 0 - margin)
        x2, y2 = self.to_local((self.width + margin), (self.height + margin))
        c1 = self.get_latlon_at(x1, y1)
        c2 = self.get_latlon_at(x2, y2)
        return Bbox((c1.lat, c1.lon, c2.lat, c2.lon))

    bbox = AliasProperty(get_bbox, None, bind=["lat", "lon", "_zoom"])

    def unload(self):
        """Unload the view and all the layers.
        It also cancel all the remaining downloads.
        """
        self.remove_all_tiles()

    def get_window_xy_from(self, lat, lon, zoom):
        """Returns the x/y position in the widget absolute coordinates
        from a lat/lon"""
        scale = self.scale
        vx, vy = self.viewport_pos
        ms = self.map_source
        x = ms.get_x(zoom, lon) - vx
        y = ms.get_y(zoom, lat) - vy
        x *= scale
        y *= scale
        x = x + self.pos[0]
        y = y + self.pos[1]
        return x, y

    def center_on(self, *args):
        """Center the map on the coordinate :class:`Coordinate`, or a (lat, lon)
        """
        map_source = self.map_source
        zoom = self._zoom

        if len(args) == 1 and isinstance(args[0], Coordinate):
            coord = args[0]
            lat = coord.lat
            lon = coord.lon
        elif len(args) == 2:
            lat, lon = args
        else:
            raise Exception("Invalid argument for center_on")
        lon = clamp(lon, MIN_LONGITUDE, MAX_LONGITUDE)
        lat = clamp(lat, MIN_LATITUDE, MAX_LATITUDE)
        scale = self._scatter.scale
        x = map_source.get_x(zoom, lon) - self.center_x / scale
        y = map_source.get_y(zoom, lat) - self.center_y / scale
        self.delta_x = -x
        self.delta_y = -y
        self.lon = lon
        self.lat = lat
        self._scatter.pos = 0, 0
        self.trigger_update(True)

    def set_zoom_at(self, zoom, x, y, scale=None):
        """Sets the zoom level, leaving the (x, y) at the exact same point
        in the view.
        """
        zoom = clamp(
            zoom, self.map_source.get_min_zoom(), self.map_source.get_max_zoom()
        )
        if int(zoom) == int(self._zoom):
            if scale is None:
                return
            elif scale == self.scale:
                return
        scale = scale or 1.0

        # first, rescale the scatter
        scatter = self._scatter
        scale = clamp(scale, scatter.scale_min, scatter.scale_max)
        rescale = scale * 1.0 / scatter.scale
        scatter.apply_transform(
            Matrix().scale(rescale, rescale, rescale),
            post_multiply=True,
            anchor=scatter.to_local(x, y),
        )

        # adjust position if the zoom changed
        c1 = self.map_source.get_col_count(self._zoom)
        c2 = self.map_source.get_col_count(zoom)
        if c1 != c2:
            f = float(c2) / float(c1)
            self.delta_x = scatter.x + self.delta_x * f
            self.delta_y = scatter.y + self.delta_y * f
            # back to 0 every time
            scatter.apply_transform(
                Matrix().translate(-scatter.x, -scatter.y, 0), post_multiply=True
            )

        # avoid triggering zoom changes.
        self._zoom = zoom
        self.zoom = self._zoom

    def on_zoom(self, instance, zoom):
        if zoom == self._zoom:
            return
        x = self.map_source.get_x(zoom, self.lon) - self.delta_x
        y = self.map_source.get_y(zoom, self.lat) - self.delta_y
        self.set_zoom_at(zoom, x, y)
        self.center_on(self.lat, self.lon)

    def get_latlon_at(self, x, y, zoom=None):
        """Return the current :class:`Coordinate` within the (x, y) widget
        coordinate.
        """
        if zoom is None:
            zoom = self._zoom
        vx, vy = self.viewport_pos
        scale = self._scale
        return Coordinate(
            lat=self.map_source.get_lat(zoom, y / scale + vy),
            lon=self.map_source.get_lon(zoom, x / scale + vx),
        )

    def add_marker(self, marker, layer=None):
        """Add a marker into the layer. If layer is None, it will be added in
        the default marker layer. If there is no default marker layer, a new
        one will be automatically created
        """
        if layer is None:
            if not self._default_marker_layer:
                layer = MarkerMapLayer()
                self.add_layer(layer)
            else:
                layer = self._default_marker_layer
        layer.add_widget(marker)
        layer.set_marker_position(self, marker)

    def remove_marker(self, marker):
        """Remove a marker from its layer
        """
        marker.detach()

    def add_layer(self, layer, mode="window"):
        """Add a new layer to update at the same time the base tile layer.
        mode can be either "scatter" or "window". If "scatter", it means the
        layer will be within the scatter transformation. It's perfect if you
        want to display path / shape, but not for text.
        If "window", it will have no transformation. You need to position the
        widget yourself: think as Z-sprite / billboard.
        Defaults to "window".
        """
        assert mode in ("scatter", "window")
        if self._default_marker_layer is None and isinstance(layer, MarkerMapLayer):
            self._default_marker_layer = layer
        self._layers.append(layer)
        c = self.canvas
        if mode == "scatter":
            self.canvas = self.canvas_layers
        else:
            self.canvas = self.canvas_layers_out
        layer.canvas_parent = self.canvas
        super().add_widget(layer)
        self.canvas = c

    def remove_layer(self, layer):
        """Remove the layer
        """
        c = self.canvas
        self._layers.remove(layer)
        self.canvas = layer.canvas_parent
        super().remove_widget(layer)
        self.canvas = c

    def sync_to(self, other):
        """Reflect the lat/lon/zoom of the other MapView to the current one.
        """
        if self._zoom != other._zoom:
            self.set_zoom_at(other._zoom, *self.center)
        self.center_on(other.get_latlon_at(*self.center))

    # Private API

    def __init__(self, **kwargs):
        from kivy.base import EventLoop

        EventLoop.ensure_window()
        self._invalid_scale = True
        self._tiles = []
        self._tiles_bg = []
        self._tilemap = {}
        self._layers = []
        self._default_marker_layer = None
        self._need_redraw_all = False
        self._transform_lock = False
        self.trigger_update(True)
        self.canvas = Canvas()
        self._scatter = MapViewScatter()
        self.add_widget(self._scatter)
        with self._scatter.canvas:
            self.canvas_map = Canvas()
            self.canvas_layers = Canvas()
        with self.canvas:
            self.canvas_layers_out = Canvas()
        self._scale_target_anim = False
        self._scale_target = 1.0
        self._touch_count = 0
        self.map_source.cache_dir = self.cache_dir
        Clock.schedule_interval(self._animate_color, 1 / 60.0)
        self.lat = kwargs.get("lat", self.lat)
        self.lon = kwargs.get("lon", self.lon)
        super().__init__(**kwargs)

    def _animate_color(self, dt):
        # fast path
        d = self.animation_duration
        if d == 0:
            for tile in self._tiles:
                if tile.state == "need-animation":
                    tile.g_color.a = 1.0
                    tile.state = "animated"
            for tile in self._tiles_bg:
                if tile.state == "need-animation":
                    tile.g_color.a = 1.0
                    tile.state = "animated"
        else:
            d = d / 1000.0
            for tile in self._tiles:
                if tile.state != "need-animation":
                    continue
                tile.g_color.a += dt / d
                if tile.g_color.a >= 1:
                    tile.state = "animated"
            for tile in self._tiles_bg:
                if tile.state != "need-animation":
                    continue
                tile.g_color.a += dt / d
                if tile.g_color.a >= 1:
                    tile.state = "animated"

    def add_widget(self, widget):
        if isinstance(widget, MapMarker):
            self.add_marker(widget)
        elif isinstance(widget, MapLayer):
            self.add_layer(widget)
        else:
            super().add_widget(widget)

    def remove_widget(self, widget):
        if isinstance(widget, MapMarker):
            self.remove_marker(widget)
        elif isinstance(widget, MapLayer):
            self.remove_layer(widget)
        else:
            super().remove_widget(widget)

    def on_map_relocated(self, zoom, coord):
        pass

    def animated_diff_scale_at(self, d, x, y):
        self._scale_target_time = 1.0
        self._scale_target_pos = x, y
        if self._scale_target_anim is False:
            self._scale_target_anim = True
            self._scale_target = d
        else:
            self._scale_target += d
        Clock.unschedule(self._animate_scale)
        Clock.schedule_interval(self._animate_scale, 1 / 60.0)

    def _animate_scale(self, dt):
        diff = self._scale_target / 3.0
        if abs(diff) < 0.01:
            diff = self._scale_target
            self._scale_target = 0
        else:
            self._scale_target -= diff
        self._scale_target_time -= dt
        self.diff_scale_at(diff, *self._scale_target_pos)
        ret = self._scale_target != 0
        if not ret:
            self._pause = False
        return ret

    def diff_scale_at(self, d, x, y):
        scatter = self._scatter
        scale = scatter.scale * (2 ** d)
        self.scale_at(scale, x, y)

    def scale_at(self, scale, x, y):
        scatter = self._scatter
        scale = clamp(scale, scatter.scale_min, scatter.scale_max)
        rescale = scale * 1.0 / scatter.scale
        scatter.apply_transform(
            Matrix().scale(rescale, rescale, rescale),
            post_multiply=True,
            anchor=scatter.to_local(x, y),
        )

    def on_touch_down(self, touch):
        if not self.collide_point(*touch.pos):
            return
        if self.pause_on_action:
            self._pause = True
        if "button" in touch.profile and touch.button in ("scrolldown", "scrollup"):
            d = 1 if touch.button == "scrollup" else -1
            self.animated_diff_scale_at(d, *touch.pos)
            return True
        elif touch.is_double_tap and self.double_tap_zoom:
            self.animated_diff_scale_at(1, *touch.pos)
            return True
        touch.grab(self)
        self._touch_count += 1
        if self._touch_count == 1:
            self._touch_zoom = (self.zoom, self._scale)
        return super().on_touch_down(touch)

    def on_touch_up(self, touch):
        if touch.grab_current == self:
            touch.ungrab(self)
            self._touch_count -= 1
            if self._touch_count == 0:
                # animate to the closest zoom
                zoom, scale = self._touch_zoom
                cur_zoom = self.zoom
                cur_scale = self._scale
                if cur_zoom < zoom or cur_scale < scale:
                    self.animated_diff_scale_at(1.0 - cur_scale, *touch.pos)
                elif cur_zoom > zoom or cur_scale > scale:
                    self.animated_diff_scale_at(2.0 - cur_scale, *touch.pos)
                self._pause = False
            return True
        return super().on_touch_up(touch)

    def on_transform(self, *args):
        self._invalid_scale = True
        if self._transform_lock:
            return
        self._transform_lock = True
        # recalculate viewport
        map_source = self.map_source
        zoom = self._zoom
        scatter = self._scatter
        scale = scatter.scale
        if scale >= 2.0:
            zoom += 1
            scale /= 2.0
        elif scale < 1:
            zoom -= 1
            scale *= 2.0
        zoom = clamp(zoom, map_source.min_zoom, map_source.max_zoom)
        if zoom != self._zoom:
            self.set_zoom_at(zoom, scatter.x, scatter.y, scale=scale)
            self.trigger_update(True)
        else:
            if zoom == map_source.min_zoom and scatter.scale < 1.0:
                scatter.scale = 1.0
                self.trigger_update(True)
            else:
                self.trigger_update(False)

        if map_source.bounds:
            self._apply_bounds()
        self._transform_lock = False
        self._scale = self._scatter.scale

    def _apply_bounds(self):
        # if the map_source have any constraints, apply them here.
        map_source = self.map_source
        zoom = self._zoom
        min_lon, min_lat, max_lon, max_lat = map_source.bounds
        xmin = map_source.get_x(zoom, min_lon)
        xmax = map_source.get_x(zoom, max_lon)
        ymin = map_source.get_y(zoom, min_lat)
        ymax = map_source.get_y(zoom, max_lat)

        dx = self.delta_x
        dy = self.delta_y
        oxmin, oymin = self._scatter.to_local(self.x, self.y)
        oxmax, oymax = self._scatter.to_local(self.right, self.top)
        s = self._scale
        cxmin = oxmin - dx
        if cxmin < xmin:
            self._scatter.x += (cxmin - xmin) * s
        cymin = oymin - dy
        if cymin < ymin:
            self._scatter.y += (cymin - ymin) * s
        cxmax = oxmax - dx
        if cxmax > xmax:
            self._scatter.x -= (xmax - cxmax) * s
        cymax = oymax - dy
        if cymax > ymax:
            self._scatter.y -= (ymax - cymax) * s

    def on__pause(self, instance, value):
        if not value:
            self.trigger_update(True)

    def trigger_update(self, full):
        self._need_redraw_full = full or self._need_redraw_full
        Clock.unschedule(self.do_update)
        Clock.schedule_once(self.do_update, -1)

    def do_update(self, dt):
        zoom = self._zoom
        scale = self._scale
        self.lon = self.map_source.get_lon(
            zoom, (self.center_x - self._scatter.x) / scale - self.delta_x
        )
        self.lat = self.map_source.get_lat(
            zoom, (self.center_y - self._scatter.y) / scale - self.delta_y
        )
        self.dispatch("on_map_relocated", zoom, Coordinate(self.lon, self.lat))
        for layer in self._layers:
            layer.reposition()

        if self._need_redraw_full:
            self._need_redraw_full = False
            self.move_tiles_to_background()
            self.load_visible_tiles()
        else:
            self.load_visible_tiles()

    def bbox_for_zoom(self, vx, vy, w, h, zoom):
        # return a tile-bbox for the zoom
        map_source = self.map_source
        size = map_source.dp_tile_size
        scale = self._scale

        max_x_end = map_source.get_col_count(zoom)
        max_y_end = map_source.get_row_count(zoom)

        x_count = int(ceil(w / scale / float(size))) + 1
        y_count = int(ceil(h / scale / float(size))) + 1

        tile_x_first = int(clamp(vx / float(size), 0, max_x_end))
        tile_y_first = int(clamp(vy / float(size), 0, max_y_end))
        tile_x_last = tile_x_first + x_count
        tile_y_last = tile_y_first + y_count
        tile_x_last = int(clamp(tile_x_last, tile_x_first, max_x_end))
        tile_y_last = int(clamp(tile_y_last, tile_y_first, max_y_end))

        x_count = tile_x_last - tile_x_first
        y_count = tile_y_last - tile_y_first
        return (tile_x_first, tile_y_first, tile_x_last, tile_y_last, x_count, y_count)

    def load_visible_tiles(self):
        map_source = self.map_source
        vx, vy = self.viewport_pos
        zoom = self._zoom
        dirs = [0, 1, 0, -1, 0]
        bbox_for_zoom = self.bbox_for_zoom
        size = map_source.dp_tile_size

        (
            tile_x_first,
            tile_y_first,
            tile_x_last,
            tile_y_last,
            x_count,
            y_count,
        ) = bbox_for_zoom(vx, vy, self.width, self.height, zoom)

        # Adjust tiles behind us
        for tile in self._tiles_bg[:]:
            tile_x = tile.tile_x
            tile_y = tile.tile_y

            f = 2 ** (zoom - tile.zoom)
            w = self.width / f
            h = self.height / f
            (
                btile_x_first,
                btile_y_first,
                btile_x_last,
                btile_y_last,
                _,
                _,
            ) = bbox_for_zoom(vx / f, vy / f, w, h, tile.zoom)

            if (
                tile_x < btile_x_first
                or tile_x >= btile_x_last
                or tile_y < btile_y_first
                or tile_y >= btile_y_last
            ):
                tile.state = "done"
                self._tiles_bg.remove(tile)
                self.canvas_map.before.remove(tile.g_color)
                self.canvas_map.before.remove(tile)
                continue

            tsize = size * f
            tile.size = tsize, tsize
            tile.pos = (tile_x * tsize + self.delta_x, tile_y * tsize + self.delta_y)

        # Get rid of old tiles first
        for tile in self._tiles[:]:
            tile_x = tile.tile_x
            tile_y = tile.tile_y

            if (
                tile_x < tile_x_first
                or tile_x >= tile_x_last
                or tile_y < tile_y_first
                or tile_y >= tile_y_last
            ):
                tile.state = "done"
                self.tile_map_set(tile_x, tile_y, False)
                self._tiles.remove(tile)
                self.canvas_map.remove(tile)
                self.canvas_map.remove(tile.g_color)
            else:
                tile.size = (size, size)
                tile.pos = (tile_x * size + self.delta_x, tile_y * size + self.delta_y)

        # Load new tiles if needed
        x = tile_x_first + x_count // 2 - 1
        y = tile_y_first + y_count // 2 - 1
        arm_max = max(x_count, y_count) + 2
        arm_size = 1
        turn = 0
        while arm_size < arm_max:
            for i in range(arm_size):
                if (
                    not self.tile_in_tile_map(x, y)
                    and y >= tile_y_first
                    and y < tile_y_last
                    and x >= tile_x_first
                    and x < tile_x_last
                ):
                    self.load_tile(x, y, size, zoom)

                x += dirs[turn % 4 + 1]
                y += dirs[turn % 4]

            if turn % 2 == 1:
                arm_size += 1

            turn += 1

    def load_tile(self, x, y, size, zoom):
        if self.tile_in_tile_map(x, y) or zoom != self._zoom:
            return
        self.load_tile_for_source(self.map_source, 1.0, size, x, y, zoom)
        # XXX do overlay support
        self.tile_map_set(x, y, True)

    def load_tile_for_source(self, map_source, opacity, size, x, y, zoom):
        tile = Tile(size=(size, size), cache_dir=self.cache_dir)
        tile.g_color = Color(1, 1, 1, 0)
        tile.tile_x = x
        tile.tile_y = y
        tile.zoom = zoom
        tile.pos = (x * size + self.delta_x, y * size + self.delta_y)
        tile.map_source = map_source
        tile.state = "loading"
        if not self._pause:
            map_source.fill_tile(tile)
        self.canvas_map.add(tile.g_color)
        self.canvas_map.add(tile)
        self._tiles.append(tile)

    def move_tiles_to_background(self):
        # remove all the tiles of the main map to the background map
        # retain only the one who are on the current zoom level
        # for all the tile in the background, stop the download if not yet started.
        zoom = self._zoom
        tiles = self._tiles
        btiles = self._tiles_bg
        canvas_map = self.canvas_map
        tile_size = self.map_source.tile_size

        # move all tiles to background
        while tiles:
            tile = tiles.pop()
            if tile.state == "loading":
                tile.state = "done"
                continue
            btiles.append(tile)

        # clear the canvas
        canvas_map.clear()
        canvas_map.before.clear()
        self._tilemap = {}

        # unsure if it's really needed, i personnally didn't get issues right now
        # btiles.sort(key=lambda z: -z.zoom)

        # add all the btiles into the back canvas.
        # except for the tiles that are owned by the current zoom level
        for tile in btiles[:]:
            if tile.zoom == zoom:
                btiles.remove(tile)
                tiles.append(tile)
                tile.size = tile_size, tile_size
                canvas_map.add(tile.g_color)
                canvas_map.add(tile)
                self.tile_map_set(tile.tile_x, tile.tile_y, True)
                continue
            canvas_map.before.add(tile.g_color)
            canvas_map.before.add(tile)

    def remove_all_tiles(self):
        # clear the map of all tiles.
        self.canvas_map.clear()
        self.canvas_map.before.clear()
        for tile in self._tiles:
            tile.state = "done"
        del self._tiles[:]
        del self._tiles_bg[:]
        self._tilemap = {}

    def tile_map_set(self, tile_x, tile_y, value):
        key = tile_y * self.map_source.get_col_count(self._zoom) + tile_x
        if value:
            self._tilemap[key] = value
        else:
            self._tilemap.pop(key, None)

    def tile_in_tile_map(self, tile_x, tile_y):
        key = tile_y * self.map_source.get_col_count(self._zoom) + tile_x
        return key in self._tilemap

    def on_size(self, instance, size):
        for layer in self._layers:
            layer.size = size
        self.center_on(self.lat, self.lon)
        self.trigger_update(True)

    def on_pos(self, instance, pos):
        self.center_on(self.lat, self.lon)
        self.trigger_update(True)

    def on_map_source(self, instance, source):
        if isinstance(source, string_types):
            self.map_source = MapSource.from_provider(source)
        elif isinstance(source, (tuple, list)):
            cache_key, min_zoom, max_zoom, url, attribution, options = source
            self.map_source = MapSource(
                url=url,
                cache_key=cache_key,
                min_zoom=min_zoom,
                max_zoom=max_zoom,
                attribution=attribution,
                cache_dir=self.cache_dir,
                **options
            )
        elif isinstance(source, MapSource):
            self.map_source = source
        else:
            raise Exception("Invalid map source provider")
        self.zoom = clamp(self.zoom, self.map_source.min_zoom, self.map_source.max_zoom)
        self.remove_all_tiles()
        self.trigger_update(True)



import sys
import re
from kivy.app import App
from kivy.logger import Logger
from kivy.uix.image import Image
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.textinput import TextInput
from kivy.uix.button import Button
from kivy.uix.dropdown import DropDown
from kivy.uix.switch import Switch
from kivy.metrics import sp,dp
from kivy.app import App
from kivy.properties import BooleanProperty, ListProperty, \
			NumericProperty

from ..threadcall import HttpClient
from ..utils import CSize, set_widget_height, set_widget_width
from ..widget_css import WidgetCSS, get_css

class BoolInput(Switch):
	def __init__(self,csscls='input', **kw):
		super().__init__()
		self.register_event_type('on_changed')
		self.bind(active=on_active)

	def on_changed(self,v=None):
		pass

	def on_active(self,t,v):
		self.dispatch('on_changed',v)
		
	def getValue(self):
		return self.active

	def setValue(self,v):
		self.active = v
	
class StrInput(WidgetCSS, TextInput):
	length = NumericProperty(None)
	def __init__(self, **kv):
		if kv is None:
			kv = {}
		a = {
			"allow_copy":True,
			"password":False,
			"multiline":False,
			"halign":"left",
			"hint_text":"test",
			"text_language":"zh_CN",
			"font_size":CSize(1),
			"write_tab":False
		}
		a.update(kv)
		super(StrInput, self).__init__(**a)
		self.text = self.old_value = ''
		self.register_event_type('on_changed')
		self.bind(on_text_validate=self.checkChange)

	def on_bgcolor(self, *args):
		self.background_color = self.bgcolor

	def on_fgcolor(self, *args):
		self.foreground_color = self.fgcolor

	def on_changed(self,v=None):
		pass
		
	def checkChange(self,o,v=None):
		v = self.getValue()
		if v != self.old_value:
			self.old_value = v
			self.dispatch('on_changed',v)

	def insert_text(self, substring, from_undo=False):
		if self.length:
			if len(self.text) + len(substring) > self.length:
				return None
		ret = super().insert_text(substring, from_undo=from_undo)
		# ret is None
		return ret

	def getValue(self):
		return self.text

	def setValue(self,v):
		if v is None:
			v = ''
		self.text = str(v)
		self.old_value = self.text
		
class Password(StrInput):
	def __init__(self, **kw):
		super().__init__(**kw)
		self.password = True

class IntegerInput(StrInput):
	def __init__(self,**kw):
		a = {}
		a.update(kw)
		a['halign'] = 'right'
		super().__init__(**a)
		self.input_filter = 'int'

	def on_focus(self,t,v):
		self.cursor = (0,len(self.text))

	def getValue(self):
		try:
			return int(self.text)
		except:
			pass
		return None

class FloatInput(IntegerInput):
	dec = NumericProperty(2)
	def __init__(self,**kw):
		super(FloatInput, self).__init__(**kw)
		self.input_filter = 'float'

	def getValue(self):
		try:
			return float(self.text)
		except:
			pass
		return None

class MyDropDown(DropDown):
	def __init__(self,csscls='input', **kw):
		super(MyDropDown,self).__init__()
		self.options = kw
		self.textField = kw.get('textField','text')
		self.valueField = kw.get('valueField','value')
		self.loadSelectItems()

	def loadSelectItems(self):
		if self.options.get('url') is not None:
			self.url = self.options.get('url')
			self.setDataByUrl(self.url)
		else:
			self.si_data = self.options.get('data')
			if self.si_data:
				self.setData(self.si_data)
		self.bind(on_select=lambda instance, x: self.selectfunc(x))

	def selectfunc(self,v):
		f = self.options.get('on_select')
		if f is not None:
			return f(v)
			
	def getTextByValue(self,v):
		for d in self.si_data:
			if d.get(self.valueField) == v:
				r = d.get(self.textField)
				return r if r else v
		return str(v)
		
	def getValueByText(self,v):
		for d in self.si_data:
			if d.get(self.textField) == v:
				r = d.get(self.valueField)
				return r
		return None
		
	def setData(self,data):
		self.si_data = data
		self.clear_widgets()
		h = self.options.get('height',2.5)
		a = {}
		a['size_hint_y'] = None
		a['height'] = CSize(2)
		a['font_size'] = CSize(1)
		for d in data:
			v = d.get(self.valueField, None)
			t = d.get(self.textField, None)
			dd = (v,t or v)
			b = Button(text=dd[1],**a)
			setattr(b,'kw_data',dd)
			b.bind(on_release=lambda btn: self.select(btn.kw_data))
			self.add_widget(b)
			
	def setDataByUrl(self,url,kw={}):
		hc = HttpClient()
		params = self.options.get('params', {}).copy()
		params.update(kw)
		d = hc.get(url,params=params)
		self.setData(d)
			
	def showme(self,w):
		self.target = w
		self.open(w)
		
class SelectInput(BoxLayout):
	def __init__(self,csscls='input', **kw):
		a={}
		w = kw.get('width',10)
		h = kw.get('height',2.5)
		if w <= 1:
			a['size_hint_x'] = w
		else:
			a['size_hint_x'] = None
			a['width'] = CSize(w)
		if h <= 1:
			a['size_hint_y'] = h
		else:
			a['size_hint_y'] = None
			a['height'] = CSize(h)

		super(SelectInput,self).__init__(orientation='horizontal',\
				**a)
		self.si_value = ''
		self.tinp = StrInput()
		# self.tinp.readonly = True
		newkw = {}
		newkw.update(kw)
		# newkw.update({'on_select':self.setData})
		self.dropdown = MyDropDown(**newkw)
		if kw.get('value'):
			self.si_data = kw.get('value')
			self.text = self.dropdown.getTextByValue(self.si_data)
		else:
			self.si_data = ''
			self.text = ''
		self.dropdown.bind(on_select=self.dropdown_select)
		self.tinp.text = self.text
		self.add_widget(self.tinp)
		self.old_value = self.si_data
		self.tinp.bind(focus=self.showDropdown)
		self.register_event_type('on_changed')
		
	def dropdown_select(self, o, d):
		self.setData(d)

	def set_selectable_data(self, data):
		self.dropdown.setData(data)

	def on_changed(self,v=None):
		pass

	def showDropdown(self,instance,yn):
		# if self.collide_point(*touch.pos):
		if yn:
			self.tinp.focus = False
			self.dropdown.showme(self)
			self.old_value = self.getValue()
		
	def setData(self,d):
		self.si_value = d[0]
		self.tinp.text = d[1]
		v = self.getValue()
		if v != self.old_value:
			self.dispatch('on_changed',v)

	def setValue(self,v):
		self.si_value = v
		self.tinp.text = self.dropdown.getTextByValue(v)
		
	def getValue(self):
		return self.si_value
		
if __name__ == '__main__':
	from kivy.app import App
	from kivy.uix.boxlayout import BoxLayout
	class MyApp(App):
		def build(self):
			root = BoxLayout(orientation='vertical')
			x = SelectInput(width=CSize(15),value='1',data=[{'value':'1','text':'ban'},{'value':'0','text':'nu'}],textField='text',valueField='value')
			root.add_widget(x)
			b = Button(text='drop', size_hint=(None, None))
			root.add_widget(b)
			dd = MyDropDown(width=CSize(15),value='1',data=[{'value':'1','text':'nan'},{'value':'0','text':'nu'}],textField='text',valueField='value',on_select=x.setData)
			b.bind(on_release=dd.showme)
			return root
	MyApp().run()
			
		



import traceback

from kivy.clock import Clock
from kivy.factory import Factory
from kivy.properties import ListProperty, BooleanProperty, \
			StringProperty, DictProperty, NumericProperty
from kivy.uix.treeview import TreeView, TreeViewNode, TreeViewLabel
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.image import AsyncImage

from .threadcall import HttpClient
from .scrollpanel import ScrollPanel
from .clickable import SingleCheckBox
from .baseWidget import Text
from .utils import CSize

from appPublic.registerfunction import getRegisterFunctionByName

class TreeViewComplexNode(BoxLayout, TreeViewLabel):
	otext = StringProperty(None)
	checkbox = BooleanProperty(False)
	icon = StringProperty(None)
	data = DictProperty(None)
	def __init__(self, **kw):
		super(TreeViewComplexNode, self).__init__(**kw)
		self.orientation = 'horizontal'
		self.size_hint_x = None
		if self.checkbox:
			cb = SingleCheckBox(size_hint=(None,None))
			cb.bind(on_press=self.set_checked)
			self.add_widget(cb)
		if self.icon:
			img = AsyncImage(source=self.icon, size_hint=(None,None))
			img.size = CSize(1,1)
			self.add_widget(img)
		txt = Text(otext=self.otext, i18n=True)
		txt.texture_update()
		txt.size_hint = (None, None)
		txt.size = txt.texture_size
		self.add_widget(txt)
		
	def set_checked(self, o):
		if not self.data:
			return
		if o.state():
			self.data['checked'] = True
		else:
			self.data['checked'] = False
			
class Hierarchy(ScrollPanel):
	url = StringProperty(None)
	params = DictProperty(None)
	method = StringProperty('get')
	idField = StringProperty(None)
	textField = StringProperty(None)
	data = ListProperty(None)
	checkbox = BooleanProperty(False)
	icon = StringProperty(None)
	single_expand = BooleanProperty(False)
	def __init__(self, **kw):
		self.register_event_type('on_press')
		self.tree = TreeView(hide_root=True)
		self.tree.size_hint = (None, None)
		self.tree.bind(on_node_expand=self.check_load_subnodes)
		self.tree.bind(selected_node=self.node_selected)
		super(Hierarchy, self).__init__(inner=self.tree, **kw)
		if self.url:
			self.data = self.get_remote_data()

	def on_press(self, node):
		print('selected node=', node)

	def node_selected(self, o, v):
		if self.tree.selected_node:
			self.dispatch('on_press', o.selected_node)

	def collapse_others(self, node):
		for n in self.tree.iterate_open_nodes(node=node.parent_node):
			if n != node and n !=node.parent_node and n.is_open:
				self.tree.toggle_node(n)

	def check_load_subnodes(self, treeview, node):
		if self.single_expand:
			self.collapse_others(node)
		if node.is_loaded:
			return
		if not self.url:
			node.is_loaded = True
			return
		data = self.get_remote_data(node)
		for d in data:
			self.create_new_node(d, node)
		node.is_loaded = True

	def create_new_node(self, data, node=None):
		n = TreeViewComplexNode(otext=data[self.textField],
			checkbox=self.checkbox,
			icon=data.get('icon') or self.icon
		)
		n.data = data
		n.width = self.tree.indent_start + \
						self.tree.indent_level * n.level \
						+ sum([i.width for i in n.children])
		if node:
			self.tree.add_node(n, node)
		else:
			self.tree.add_node(n)
		children = data.get('children')
		if not children:
			return
		for c in children:
			self.create_new_node(c, n)
		n.is_loaded = True
		
	def get_remote_data(self, node=None):
		if not self.url:
			return
		hc = HttpClient()
		params = self.params.copy() if self.params else {}
		if node:
			params['id'] = node.data[self.idField]
		print('params=', params)
		d = hc(self.url, method=self.method, params=params)
		if isinstance(d, list):
			return d
		if isinstance(d, dict):
			return d['rows']
		return None

	def on_data(self, o, d=None):
		if not self.data:
			return
		for d in self.data:
			self.create_new_node(d)
			
class Menu(Hierarchy):
	target = StringProperty(None)
	def __init__(self, **kw):
		self.target_w = None
		super(Menu, self).__init__(**kw)

	def node_selected(self, o, v):
		if not v.is_leaf:
			self.tree.toggle_node(v)
			return
		super().node_selected(o, v)

	def on_press(self, node):
		self.tree.deselect_node()
		data = {}
		dw = node.data.get('datawidget')
		if dw:
			data_widget = Factory.Blocks.getWidgetById(dw)
			if data_widget:
				vn = node.data.get('datamethod', 'getValue')
				if hasattr(data_widget, vn):
					f = getattr(data_widget, vn)
					data = f()
					if not isinstance(data, dict):
						data = {}

		url = node.data.get('url')
		target = Factory.Blocks.getWidgetById(node.data.get('target',self.target),self)
		if url:
			params = node.data.get('params',{})
			params.update(data)
			blocks = Factory.Blocks()
			desc = {
				"widgettype":"urlwidget",
				"options":{
					"url":url,
					"params":params
				}
			}
			w = blocks.widgetBuild(desc)
			if w and target:
				target.add_widget(w)
			return 

		rfname = node.data.get('rfname')
		if rfname:
			f = getRegisterFunctionByName(rfname)
			if f:
				f(self, **data)
			return
		
		script = node.data.get('script')
		if script:
			target_name = node.data.get('target', self.target)
			target =  Factory.Blocks.getWidgetById(target_name, self)
			data.update({'self':target})
			if target:
				eval(script,data)
			return
		
Factory.register('Hierarchy', Hierarchy)
Factory.register('Menu', Menu)



"""
BoxViewer options:
{
	"dataloader":{
		"url",
		"method",
		"params"
		"filter":{
		}
	}
	"boxwidth",
	"boxheight",
	"viewer"
	"toolbar":{
	}
}
"""
from functools import partial

from traceback import print_exc
from functools import partial
from appPublic.dictExt import dictExtend
from kivy.app import App
from kivy.factory import Factory
from kivy.clock import Clock
from kivy.utils import platform
from kivy.uix.boxlayout import BoxLayout
from kivy.graphics import Fbo, Color, Rectangle
from kivy.properties import NumericProperty, StringProperty, DictProperty
from .responsivelayout import VResponsiveLayout
from .toolbar import Toolbar
from .paging import Paging, RelatedLoader
from .utils import CSize, SUPER
from .ready import WidgetReady
from .baseWidget import VBox


class BoxViewer(VBox):
	toolbar = DictProperty(None)
	dataloader = DictProperty(None)
	boxwidth = NumericProperty(None)
	boxwidth_c = NumericProperty(None)
	boxheight = NumericProperty(None)
	boxheight_c = NumericProperty(None)
	viewer = DictProperty(None)
	viewer_url = StringProperty(None)
	viewer_css = StringProperty('viewer_css')
	def __init__(self, **options):
		self.subwidgets = []
		self.toolbar_w = None
		self._dataloader = None
		self.initflag = False
		SUPER(BoxViewer, self, options)
		self.selected_data = None
		self.viewContainer = VResponsiveLayout(box_width=self.box_width)
		if self.toolbar:
			self.toolbar_w = Toolbar(self.toolbar)
		lopts = self.dataloader.get('options', {}).copy()
		self._dataloader = RelatedLoader(target=self,**lopts)
		self._dataloader.bind(on_deletepage=self.deleteWidgets)
		self._dataloader.bind(on_pageloaded=self.addPageWidgets)
		self._dataloader.bind(on_newbegin=self.deleteAllWidgets)
		self.params = lopts.get('params',{})

		if self.toolbar_w:
			self.add_widget(self.toolbar_w)
		if self._dataloader.widget:
			self.add_widget(self._dataloader.widget)
			self._dataloader.widget.bind(on_submit=self.getParams)
		self.add_widget(self.viewContainer)
		self.register_event_type('on_selected')
		self.bind(size=self.resetCols,
								pos=self.resetCols)
		self.viewContainer.bind(on_scroll_stop=self.on_scroll_stop)

	def on_boxheight(self, *args):
		self.box_height = self.boxheight

	def on_boxheight_c(self, *args):
		self.box_height = CSize(self.boxheight_c)

	def on_boxwidth_c(self, *args):
		self.box_width = CSize(self.boxwidth_c)

	def on_boxwidth(self, *args):
		self.box_width = self.boxwidth

	def key_handle(self,keyinfo):
		print('keyinfo=',keyinfo,'...................')

	def getParams(self,o,p):
		self.params = p

	def deleteAllWidgets(self,o):
		self.viewContainer.clear_widgets()
		self.subwidgets = []
	
	def getShowRows(self):
		wc = self.viewContainer.get_cols()
		hc = int(self.viewContainer.height / self.box_height)
		self.show_rows = wc * hc

	def addPageWidgets(self,o,data):
		widgets = []
		recs = data['data']
		page = data['page']
		dir = data['dir']
		idx = 0
		if dir == 'up':
			recs.reverse()
			idx = -1
		with self.fboContext():
			for r in recs:
				self.showObject(widgets, r, index=idx)
		
		data['widgets'] = widgets
		data['idx'] = idx

	def add_page_delay(self, data, *args):
		recs = data['data']
		page = data['page']
		idx = data['idx']
		widgets = data['widgets']

		self._fbo = Fbo(size=self.size)
		with self._fbo:
			self._background_color = Color(0,0,0,1)
			self._background_rect = Rectangle(size=self.size)
		for r in recs:
			self.showObject(widgets, r, index=idx)
		with self.canvas:
			self._fbo_rect = Rectangle(size=self.size,
								texture=self._fbo.texture)

		self.subwidgets += widgets

		self._dataloader.bufferObjects(page, widgets)
		x = self._dataloader.getLocater()
		self.locater(x)

	def loadData(self, **kw):
		self.params = kw
		self.deleteAllWidgets(None)
		self._dataloader.loadPage(1)
		self.initflag = True

	def deleteWidgets(self,o,data):
		for w in data:
			self.viewContainer.remove_widget(w)
		self.subwidget = [i for i in self.subwidgets if i in data]

	def on_selected(self, data=None):
		pass

	def locater(self, pos):
		self.viewContainer.scroll_y = pos

	def resetCols(self,o=None, v=None):
		h = 0
		if self.toolbar:
			h += self.toolbar.height
		if self._dataloader.widget:
			h += self._dataloader.widget.height
		self.viewContainer.height = self.height - h

		self.viewContainer.setCols()
		self.getShowRows()
		if not self.initflag:
			self._dataloader.loadPage(1)
			self.initflag = True

	def showObject(self, holders, rec,index=0):
		opts = {
			"csscls":self.viewer_css
		}
		desc = {
			"widgettype":"PressableBox",
			"options":opts
		}
		blocks = Factory.Blocks()
		box = blocks.w_build(desc)
		box.setValue(rec)
		box.bind(on_press=self.select_record)
		self.viewContainer.add_widget(box,index=index)
		self.subwidgets.append(box)
		holders.append(box)
		desc = {}
		if self.viewer_url:
			desc = {
				"widgettype":"urlwidget",
				"options":{
					"params":rec,
					"method":"GET",
					"url":self.viewer_url
				}
			}
		else:
			desc = self.viewer.copy()
			if desc['widgettype'] == 'urlwidget':
				desc = dictExtend(desc,{'options':{'params':rec}})
			else:
				desc = dictExtend(desc,{'options':{'record':rec}})
		def add2box(p,o,w):
			p.add_widget(w)

		blocks = Factory.Blocks()
		blocks.bind(on_built=partial(add2box,box))
		blocks.widgetBuild(desc)

	def on_scroll_stop(self,o,v=None):
		if o.scroll_y <= 0.001:
			self._dataloader.loadNextPage()
		if o.scroll_y >= 0.999:
			self._dataloader.loadPreviousPage()

	def select_record(self,o,v=None):
		for w in self.subwidgets:
			w.box_actived = False
		o.box_actived = True
		self.selected_data = o.getValue()
		self.dispatch('on_selected',self.selected_data)

	def getValue(self):
		return self.selected_data
		d = {
			"caller":self,
			"page_rows":1,
			"page":self.selected_data['__posInSet__'],
			"dataurl":self.dataloader['options']['dataurl'],
			"params":self.params
		}
		return d



from kivy.utils import platform
import plyer
try:
	from plyer.platforms.android import activity
except:
	print('android_rotation.py:import ERROR')
	activity = None

ROTATION_0 = 0
ROTATION_90 = 1
ROTATION_180 = 2
ROTATION_270 = 3
def get_rotation():
	try:
		r = activity.getWindowManager().getDefaultDisplay().getRotation()
		print('get_rotation(): rotation=', r, type(r))
		return r
	except Exception as e:
		print('get_rotation() failed',e)
		return None



from kivy.logger import Logger
from kivy.graphics import Color, Rectangle
from kivy.uix.button import ButtonBehavior
from kivy.uix.image import AsyncImage
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.gridlayout import GridLayout
from kivy.uix.label import Label
from kivy.uix.widget import Widget
from kivy.app import App
from kivy.clock import Clock
from kivy.factory import Factory
from kivy.properties import StringProperty, DictProperty, \
			OptionProperty, ListProperty, NumericProperty, \
			BooleanProperty

from appPublic.dictObject import DictObject
from appPublic.registerfunction import RegisterFunction

from .widgetExt.scrollwidget import ScrollWidget
from .utils import *
from .ready import WidgetReady
from .color_definitions import getColors
from .baseWidget import Text, Box
from .scrollpanel import ScrollPanel
from .command_action import cmd_action

class Toolbar(ScrollPanel):
	"""
	tool has the follow attributes
	{
		"name",
		"source_on,
		"source_off,
		# source_on, source_off, or "img_src",
		"label",
		"deletable"
		"url",
		"rfname",
	}
	toolbar has follow attributes
	{
		"img_size_c":image height in charecter size
		"text_size_c":
		"toolbar_orient":"H" or "V"
		"tool_orient":"horizontal" or "vertical"
		"css_on",
		"css_off",
		"tools":
	}
	"""

	css_on = StringProperty('default')
	css_off = StringProperty('default')
	tools = ListProperty([])
	tool_orient = OptionProperty('horizontal', \
						options=['horizontal', 'vertical'])
	toolbar_orient = OptionProperty('H', options=['H', 'V'])
	img_size_c = NumericProperty(2)
	text_size_c = NumericProperty(1)
	target = StringProperty(None)
	executable = BooleanProperty(False)
	def __init__(self, **kw):
		self.w_dic = {}
		SUPER(Toolbar, self, kw)
		self.buffer = {}
		self.register_event_type('on_press')
		self.register_event_type('on_delete_tool')
		if self.toolbar_orient == 'H':
			self._inner.orientation = 'horizontal'
		else:
			self._inner.orientation = 'vertical'
		self.clear_widgets()
		for t in self.tools:
			self.add_tool(t)
		
		self.bar_width = 0
		if self.toolbar_orient == 'H':
			self.do_scroll_y = False
		else:
			self.do_scroll_x = False

	def on_children_size(self, o, size):
		self.on_size(None, None)

	def on_size(self, o, size):
		if self.toolbar_orient == 'H':
			self.size_hint_x = 1
			self.size_hint_y = None
			if len(self.w_dic.keys()) > 0:
				self.height = max([w.height for w in self.w_dic.keys()])
		else:
			self.size_hint_x = None
			self.size_hint_y = 1
			if len(self.w_dic.keys()) > 0:
				self.width = max([w.width for w in self.w_dic.keys() ])
		
	def add_tool(self, t):
		label = t.get('label', t.get('name', None))
		source_on = t.get('source_on', t.get('img_src',None))
		source_off = t.get('source_off', t.get('img_src', None))
		ToggleIconText = Factory.ToggleIconText
		ToggleText = Factory.ToggleText
		ToggleImage = Factory.ToggleImage
		ClickableImage = Factory.ClickableImage
		if label and source_on:
			w = ToggleIconText(css_on=self.css_on,
							css_off=self.css_off,
							text=label,
							size_hint = (None, None),
							fontsize=self.text_size_c,
							source_on=source_on,
							source_off=source_off,
							orientation=self.tool_orient,
							img_kw={
								"size_hint":(None, None),
								"height":CSize(self.img_size_c),
								"width":CSize(self.img_size_c)
							}
			)
		elif label:
			w = ToggleText(css_on=self.css_on,
							css_off=self.css_off,
							size_hint = (None, None),
							orientation=self.tool_orient,
							fontsize=CSize(self.text_size_c),
							text=label)
		elif source_on:
			w = ToggleImage( source_on=source_on,
							source_off=source_off,
							size_hint = (None, None),
							orientation=self.tool_orient,
							img_kw={
								"size_hint":(None, None),
								"height":CSize(self.img_size_c),
								"width":CSize(self.img_size_c)
							}
			)

		w.widget_id = t['name']
		if t.get('deletable', False):
			x = ClickableImage(source=blockImage('delete.png'),
							size_hint=(None,None),
							size=CSize(1,1)
						)
			x.data = (w, t)
			x.bind(on_press=self.delete_tool)
			w.add_widget(x)

		if w:
			self.add_widget(w)
			self.w_dic[w] = t.copy()
			self.w_dic[w]['widget'] = w
			w.select(False)
			w.bind(size=self.on_children_size)
			w.bind(on_press=self.tool_press)
			w.bind(minimum_width=w.setter('width'))
			w.bind(minimum_height=w.setter('height'))

	def delete_tool(self, o):
		w, t = o.data
		self.del_tool(w)
		self.dispatch('on_delete_tool', o.data)

	def on_delete_tool(self, o, d=None):
		pass

	def del_tool(self, w):
		self.remove_widget(w)
		self.w_dic = {k:v for k,v in self.w_dic.copy().items() \
						if k != w }
		
	def on_press(self, o):
		pass

	def select(self,name):
		for w, v in self.w_dic.items():
			if v['name'] == name:
				w.select(True)
				self.tool_press(w)
				break

	def tool_press(self, o):
		for w,v in self.w_dic.items():
			if w == o:
				ret_v = v
				w.select(True)
			else:
				w.select(False)
		if not self.executable:
			self.dispatch('on_press', ret_v)
			return

		desc = self.w_dic[o].copy()
		if not desc.get('target'):
			desc['target'] = self.target

		cmd_action(desc, self)

class ToolPage(Box):
	"""
	ToolPage:
	{
		toolbar_size:
		tool_at
		toolbar
	}
	"""
	toolbar_size = NumericProperty(2)
	tool_at = OptionProperty('top', \
					options=['top', 'bottom', 'left', 'right'])
	toolbar = DictProperty({})
	def __init__(self, **kw):
		SUPER(ToolPage, self, kw)
		if self.tool_at in [ 'top','bottom']:
			self.orientation = 'vertical'
		else:
			self.orientation = 'horizontal'
		if not self.toolbar_size:
			self.toolbar_size = self.img_size + self.text_size + 0.3
		self.content_widgets = {}
		self.content_w = None
		self.toolbar_w = None
		self.init()
		self.toolbar_w.bind(on_delete_tool=self.delete_content_widget)
		name = self.toolbar['tools'][0]['name']
		self.toolbar_w.select(name)
	
	def get_subwidgets(self):
		children = [ i for i in self.children]
		x = [ i for i in self.content_widgets.values() ]
		return children + x

	def delete_content_widget(self, o, d):
		w, dic = d
		self.del_tool(dic['name'])

	def add_tool(self, tool):
		self.toolbar_w.add_tool(tool)

	def del_tool(self, name):
		g = self.content_widgets.copy()
		self.content_widgets = {k:v for k,v in g.items() if k!=name}

	def on_size(self,obj,size):
		if self.content_w is None:
			return
		if self.tool_at in ['top','bottom']:
			self.toolbar_w.size_hint_y = None
			self.content_w.height = self.height - self.toolbar_w.height
		else:
			self.toolbar_w.size_hint_x = None
			self.content_w.width = self.width - self.toolbar_w.width

	def init(self):
		self.content_w = BoxLayout()
		self.content_w.widget_id = 'content'
		self.toolbar_w = Toolbar(**self.toolbar)
		if self.tool_at in ['top','left']:
			self.add_widget(self.toolbar_w)
			self.add_widget(self.content_w)
		else:
			self.add_widget(self.content_w)
			self.add_widget(self.toolbar_w)
		
		self.toolbar_w.bind(on_press=self.on_press_handle)

	def build_widget(self, url):
		desc = {
			"widgettype":"urlwidget",
			"options":{
				"url":url
			}
		}
		b = Factory.Blocks()
		return b.widgetBuild(desc)

	def print_all(self):
		self.print_info(self)
		self.print_info(self.toolbar_w)
		for w in self.toolbar_w.w_dic.keys():
			self.print_info(w)
			for c in w.children:
				self.print_info(c)
		self.print_info(self.content_w)
		for c in self.content_w.children:
			self.print_info(c)

	def print_info(self, o):
		print('self.size_hint=', o.__class__, o.size_hint, 
				'self.x,y=', o.width, o.height)
		
	def on_press_handle(self, o, v):
		name = v.get('name')
		refresh = v.get('refresh', False)
		# self.print_all()

		w = self.content_widgets.get(name)
		self.content_w.clear_widgets()
		if w and not refresh:
			self.content_w.add_widget(w)
			return
		url = v.get('url')
		if url:
			w = self.build_widget(url)
			if w:
				self.content_widgets[name] = w
				self.content_w.add_widget(w)
			return
		rfname = v.get('rfname')
		if rfname:
			rf = RegisterFunction()
			f = rf.get(rfname)
			if f:
				r = f()
				if isinstance(r,Widget):
						self.content_w.add_widget(r)




import os
import sys
import time
import threading

from appPublic.sockPackage import get_free_local_addr
from kivy.utils import platform
from traceback import print_exc
from kivy.core.window import Window
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.video import Video
from kivy.uix.slider import Slider
from kivy.uix.popup import Popup
from kivy.uix.button import Button
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.filechooser import FileChooserListView
from kivy.uix.label import Label
from kivy.app import App
from kivy.clock import Clock
from kivy.logger import Logger

from kivy.properties import ObjectProperty, StringProperty, BooleanProperty, \
    NumericProperty, DictProperty, OptionProperty
from ffpyplayer.tools import set_log_callback
from .utils import *
from .paging import PageLoader
from .baseWidget import PressableImage
from .swipebehavior import SwipeBehavior

desktopOSs=[
	"win",
	"linux",
	"macosx"
]

logger_func = {'quiet': Logger.critical, 'panic': Logger.critical,
               'fatal': Logger.critical, 'error': Logger.error,
               'warning': Logger.warning, 'info': Logger.info,
               'verbose': Logger.debug, 'debug': Logger.debug}

othersplatforms=['ios','android']

class BaseVPlayer(FloatLayout):
	fullscreen = BooleanProperty(False)
	def __init__(self,vfile=None,loop=None,mute=False):
		FloatLayout.__init__(self)
		self.loop = loop
		self.muteFlg = mute
		self.register_event_type('on_source_error')
		self.register_event_type('on_next')
		self.register_event_type('on_previous')
		Window.allow_screensaver = False
		self._video = Video(allow_stretch=True)
		self.add_widget(self._video)
		self.nextdir = None
		self.playlist = []
		self.curplay = 0
		self.old_volume = 0
		self._video.bind(state=self.on_state)
		# self._video.bind(loaded=self.playVideo) no effect
		self._video.bind(position=self.positionChanged)
		if loop:
			self.eos = 'loop'
		
		# self.bind(on_swipe_down=self.previous)
		# self.bind(on_swipe_up=self.next)
		set_log_callback(self.ffplayerLog)
		if hasattr(self._video._video, '_ffplayer'):
			self.ffplayer = self._video._video._ffplayer
		if vfile is not None:
			self.setSource(vfile)

	def on_size(self,*args):
		self._video.size = self.size
		print('****on_size()****',self.pos,self.size,self._video.pos,self._video.size)
	
	def positionChanged(self,o,v):
		if self.muteFlg:
			self._video.volume = 0

	def playVideo(self,o=None,v=None):
		# print('-------------VPlayer():playVideo()')
		self._video.state = 'play'
		self.nextdir = None

	def setSource(self,s):
		self.stop()
		self.curplay = 0
		self.playlist = [s]
		self._video.source = self.playlist[self.curplay]
		Logger.info('kivyblocks: Vplayer().setSource,s=%s',s)
		self.playVideo()

	def on_source_error(self,o,v):
		Logger.info('safecorner: {} error'.format(v))

	def on_next(self,o=None, v=None):
		pass

	def on_previous(self, o=None, v=None):
		pass

	def ffplayerLog(self, msg, level):
		msg = msg.strip()
		if msg:
			logger_func[level]('yffpyplayer: {}'.format(msg))
		if level == 'error' and self._video.source in msg:
			self.dispatch('on_source_error',self,self._video.source)

	def play(self,o=None,v=None):
		if self.curplay >= 0:
			self._video.source = self.playlist[self.curplay]
			self._video.state = 'play'
		else:
			Logger.info('VPlaer: self.curpay < 0')
			
	def next(self,o=None,v=None):
		self.nextdir = 'down'
		self.stop()
		self.dispatch('on_next',self)

	def previous(self,o=None,v=None):
		self.nextdir = 'up'
		self.stop()
		self.dispatch('on_previous',self)

	def on_state(self,o=None,v=None):
		if self._video.state == 'play':
			Window.allow_screensaver = False
		else:
			Window.allow_screensaver = True
		if self._video.state == 'stop':
			Logger.info('VPlayer: state==stop,nextdir=%s',self.nextdir)
			if self.nextdir is None:
				self.dispatch('on_next',self)
			self.nextdir = None

	def on_fullscreen(self, instance, value):
		window = self.get_parent_window()
		if not window:
			Logger.warning('VideoPlayer: Cannot switch to fullscreen, '
						   'window not found.')
			if value:
				self.fullscreen = False
			return
		if not self.parent:
			Logger.warning('VideoPlayer: Cannot switch to fullscreen, '
						   'no parent.')
			if value:
				self.fullscreen = False
			return

		if value:
			Window.fullscreen = True

			self._fullscreen_state = state = {
				'parent': self.parent,
				'pos': self.pos,
				'size': self.size,
				'pos_hint': self.pos_hint,
				'size_hint': self.size_hint,
				'window_children': window.children[:]}

			if platform in desktopOSs:
				Window.maximize()
			# remove all window children
			for child in window.children[:]:
				window.remove_widget(child)

			# put the video in fullscreen
			if state['parent'] is not window:
				state['parent'].remove_widget(self)
			window.add_widget(self)

			# ensure the video widget is in 0, 0, and the size will be
			# readjusted
			self.pos = (0, 0)
			self.size = (100, 100)
			self.pos_hint = {}
			self.size_hint = (1, 1)
		else:
			Window.fullscreen = False
			state = self._fullscreen_state
			window.remove_widget(self)
			for child in state['window_children']:
				window.add_widget(child)
			self.pos_hint = state['pos_hint']
			self.size_hint = state['size_hint']
			self.pos = state['pos']
			self.size = state['size']
			if state['parent'] is not window:
				state['parent'].add_widget(self)
			if platform in desktopOSs:
				Window.restore()

	def endplay(self,btn=None):
		self._video.seek(1.0,precise=True)

	def replay(self,btn=None):
		self._video.seek(0.0,precise=True)

	def audioswitch(self,btn=None):
		x = self._video._video._ffplayer.request_channel('audio')

	def setVolume(self,obj,v):
		self._video.volume = v
		if v >= 0.01:
			self.muteFlg = False
		else:
			self.muteFlg = True

	def setPosition(self,obj,v):
		self._video.seek(v)

	def mute(self,btn=None):
		if self._video.volume > 0.001:
			self.old_volume = self._video.volume
			self._video.volume = 0.0
		else:
			self._video.volume = self.old_volume

	def stop(self):
		self._video.unload()

	def pause(self,t=None):
		if self._video.state == 'play':
			self._video.state = 'pause'
		else:
			self._video.state = 'play'

	def __del__(self):
		self.stop()
	
class Swipe_VPlayer(BaseVPlayer, SwipeBehavior):
	def __init__(self,vfile=None, loop=False, mute=False):
		BaseVPlayer.__init__(self,vfile=vfile, loop=loop, mute=mute)
		SwipeBehavior.__init__(self)
		self.bind(on_swipe_down=self.previous)
		self.bind(on_swipe_up=self.next)

class VPlayer(Swipe_VPlayer):
	def __init__(self,vfile=None, loop=False,mute=False, opbar=True):
		self.opbar = opbar
		self.menubar = None
		self._popup = None
		self.menu_status = False
		self.manualMode = False
		self.pb = None
		super().__init__(vfile=vfile,loop=loop,mute=mute)
		self._video.bind(on_touch_down=self.show_hide_menu)
	
	def totime(self,dur):
		h = dur / 3600
		m = dur % 3600 / 60
		s = dur % 60
		return '%02d:%02d:%02d' % (h,m,s)
		
	def createProgressBar(self):
		if self.pb is None:
			self.pb = BoxLayout(orientation='horizontal',
				size_hint = (0.99,None),height=CSize(1.4))
			self.curposition = Label(text='0',width=CSize(4),
				size_hint_x=None)
			self.curposition.align='right'
			self.maxposition = Label(text=self.totime(self._video.duration),
				width=CSize(4),size_hint_x=None)
			self.maxposition.align = 'left'
			self.slider = Slider(min=0, 
				max=self._video.duration, 
				value=0, 
				orientation='horizontal', 
				step=0.01)
			self.slider.bind(on_touch_down=self.enterManualMode)
			self.slider.bind(on_touch_up=self.endManualMode)
			self.manual_mode=False

			self.pb.add_widget(self.curposition)
			self.pb.add_widget(self.slider)
			self.pb.add_widget(self.maxposition)
			self.menubar.add_widget(self.pb)

	def enterManualMode(self,obj,touch):
		if not self.slider.collide_point(*touch.pos):
			return
		self.manualMode = True

	def endManualMode(self,obj,touch):
		if not self.manualMode:
			return
		if self._video.duration < 0.0001:
			return
		self._video.seek(self.slider.value/self._video.duration)
		self.manualMode = False

	def positionChanged(self,o,v):
		self.update_slider(None)

	def update_slider(self,t):
		if self._video.state != 'play':
			return
		if self.pb is None:
			return
		v = self._video.position
		if v is None:
			return
		self.curposition.text = self.totime(v)
		if not self.manualMode:
			self.slider.value = v
			self.slider.max = self._video.duration
		self.maxposition.text = self.totime(self._video.duration)

	def __del__(self):
		self.stop()

	def beforeDestroy(self):
		print('beforeDestroy() called')
		self.stop()
		return True

	def show_hide_menu(self,obj,touch):
		if not self.collide_point(*touch.pos):
			print('not inside the player',touch.pos,self.pos,self.size)
			return 

		if touch.is_double_tap:
			self.fullscreen = False if self.fullscreen else True
			print('doube_tap')
			return 

		if not self.opbar:
			return

		if not self.menubar:
			self.buildMenu()
			return

		if self.menu_status:
			self.remove_widget(self.menubar)
		else:
			self.add_widget(self.menubar)
		self.menu_status = not self.menu_status

	def buildMenu(self):
		self.menubar = BoxLayout(orientation='horizontal',
			size_hint_y=None,height=CSize(1.4))
		self.btn_pause = PressableImage(source=blockImage('pause.png'),
					size_hint=(None,None),
					size=CSize(3,3)
		)
		if self._video.state == 'pause':
			self.btn_pause.source = blockImage('play.png')
		self.btn_pause.bind(on_press=self.pause)
		self.menubar.add_widget(self.btn_pause)
		"""
		self.btn_mute = PressableImage(source=blockImage('mute.png'),
					size_hint=(None,None),
					size=CSize(3,3)
		)
		self.btn_mute.bind(on_press=self.mute)
		self.menubar.add_widget(self.btn_mute)
		"""
		btn_cut = PressableImage(source=blockImage('next.png'),
				size_hint=(None,None),
				size=CSize(3,3)
		)
		btn_cut.bind(on_press=self.endplay)
		self.menubar.add_widget(btn_cut)
		btn_replay = PressableImage(source=blockImage('replay.png'),
				size_hint=(None,None),
				size=CSize(3,3)
		)
		btn_replay.bind(on_press=self.replay)
		self.menubar.add_widget(btn_replay)

		self.btn_audioswitch = PressableImage( \
					source=blockImage('musictrack.png'),
				size_hint=(None,None),
				size=CSize(3,3)
		)
					
		self.btn_audioswitch.bind(on_press=self.audioswitch)
		self.menubar.add_widget(self.btn_audioswitch)
		slider = Slider(min=0,
                                max=1,
                                value=self._video.volume,
                                orientation='horizontal',
								size_hint=(None,None),
								height = CSize(2),
								width = CSize(10),
                                step=0.07)
		slider.bind(on_touch_up=self.setVolume)
		self.menubar.add_widget(slider)
		self.menubar.pos = CSize(0,0)
		self.createProgressBar()
		self.add_widget(self.menubar)
		self.menu_status = True

	def setVolume(self,obj,v=None):
		BaseVPlayer.setVolume(self,obj,obj.value)

	def mute(self,btn):
		BaseVPlayer.mute(self,btn)
		if self.menubar:
			if self._video.volume < 0.001:
				btn.source = blockImage('volume.png')
			else:
				btn.source = blockImage('mute.png')

	def pause(self,t=None):
		BaseVPlayer.pause(self,t)
		if self.menubar:
			if self._video.state == 'pause':
				self.btn_pause.source  = blockImage('play.png')
			else:
				self.btn_pause.source = blockImage('pause.png')

if __name__ == '__main__':
	class MyApp(App):
		def build(self):
			vf = None
			if len(sys.argv) > 1:
				vf = sys.argv[1:]
			self.player = VPlayer(vfile=vf,
				loop=True,
				can_openfile=True,
				can_move = True,
				can_cut=True,
				can_replay=True,
				can_changevolume = True
			)
			return self.player

	MyApp().run()



# -*- utf-8 -*-
"""
模板文件以wrdesc结尾
界面描述文件以uidesc结尾
"""
import os
import sys
import codecs
from kivy.app import App

from appPublic.folderUtils import ProgramPath,listFile
from appPublic.jsonConfig import getConfig

import json

class ExternalWidgetManager:
	def __init__(self):
		#self.register_root = os.path.join(ProgramPath(),'widgets')
		#self.ui_root = os.path.join(ProgramPath(),'ui')
		self.ui_root = './ui'
		self.register_root = './widgets'
		
	def loadJson(self,filepath):
		with codecs.open(filepath,'r','utf-8') as f:
			return json.load(f)
	
	def travalRegisterDesc(self,func):
		return
		for f in listFile(self.register_root,suffixs=['.wrdesc'],rescursive=True):
			desc = self.loadJson(f)
			return func(desc)
			
	def loadWidgetDesc(self,desc):
		def text2Json(d):
			j = json.loads(d)
			return j

		# print(desc)
		if desc.get('filename'):
			path = desc.get('filename')
			if path.endswith('.uidesc'):
				f = FileDataLoader()
				f.bind(on_dataloaded=text2Json)
				if path.startswith('/'):
					path = path[1:]
				fn = os.path.join(self.ui_root,path)
				return text2Json(f.loadData(fn))
			raise Exception('file error',path)

		if desc.get('url'):
			url = desc.get('url')
			headers = desc.get('headers',{})
			params = desc.get('params',{})
			app = App.get_running_app()
			resp = app.hc.sync_get(url,params=params,headers=headers)
			if resp.status_code == 200:
				d = resp.json()
				if d.get('status') == 'OK':
					return d['data']
				raise Exception('ui desc loaded failed %s' % url)

		raise Exception('ui desc loaded failed' , desc)
	



"""
{
	"widgettype":"TabsPanel",
	"options":{
		"tab_pos":"top_left"
		"tabs":[
			{
				"text":"tab1",
				"icon":"/img/hhhh.png",
				"refresh_press":Fasle,
				"content":{
					"widgettype":"urlwidegt",
					"url":"reggtY",
				}
			},
			{
			}
		]
	},
}
"""
from appPublic.uniqueID import getID

from kivy.uix.tabbedpanel import TabbedPanel, TabbedPanelItem
from kivy.clock import Clock
from kivy.factory import Factory

from .bgcolorbehavior import BGColorBehavior

class TabsPanel(BGColorBehavior, TabbedPanel):
	def __init__(self,color_level=-1,
					radius=[],
					tabs=[],
					**options):
		self.tabs_list = tabs
		TabbedPanel.__init__(self,**options)
		BGColorBehavior.__init__(self,color_level=color_level,
						radius=radius)
		Clock.schedule_once(self.add_tabs,0)

	def newname(self):
		return getID()

	def add_tab(self,name,text,desc):
		def add(o,w):
			if not hasattr(w,'widget_id'):
				w.widget_id = name
			self.add_widget(TabbedPanelItem(text=text,content=w))
		blocks = Factory.Blocks()
		blocks.bind(on_built=add)
		blocks.widgetBuild(desc)

	def add_tabs(self,*args):
		for d in self.tabs_list:
			name = d.get('name',self.newname())
			text = d['text']
			desc = d['content']
			self.add_tab(name,text,desc)

if __name__ == '__main__':
	from kivy.app import App
	pass



# coding=utf-8
"""
MBTiles provider for MapView
============================

This provider is based on .mbfiles from MapBox.
See: http://mbtiles.org/
"""

__all__ = ["MBTilesMapSource"]


import io
import sqlite3
import threading

from kivy.core.image import Image as CoreImage
from kivy.core.image import ImageLoader

from kivyblocks.mapview.downloader import Downloader
from kivyblocks.mapview.source import MapSource


class MBTilesMapSource(MapSource):
    def __init__(self, filename, **kwargs):
        super().__init__(**kwargs)
        self.filename = filename
        self.db = sqlite3.connect(filename)

        # read metadata
        c = self.db.cursor()
        metadata = dict(c.execute("SELECT * FROM metadata"))
        if metadata["format"] == "pbf":
            raise ValueError("Only raster maps are supported, not vector maps.")
        self.min_zoom = int(metadata["minzoom"])
        self.max_zoom = int(metadata["maxzoom"])
        self.attribution = metadata.get("attribution", "")
        self.bounds = bounds = None
        cx = cy = 0.0
        cz = 5
        if "bounds" in metadata:
            self.bounds = bounds = map(float, metadata["bounds"].split(","))
        if "center" in metadata:
            cx, cy, cz = map(float, metadata["center"].split(","))
        elif self.bounds:
            cx = (bounds[2] + bounds[0]) / 2.0
            cy = (bounds[3] + bounds[1]) / 2.0
            cz = self.min_zoom
        self.default_lon = cx
        self.default_lat = cy
        self.default_zoom = int(cz)
        self.projection = metadata.get("projection", "")
        self.is_xy = self.projection == "xy"

    def fill_tile(self, tile):
        if tile.state == "done":
            return
        Downloader.instance(self.cache_dir).submit(self._load_tile, tile)

    def _load_tile(self, tile):
        # global db context cannot be shared across threads.
        ctx = threading.local()
        if not hasattr(ctx, "db"):
            ctx.db = sqlite3.connect(self.filename)

        # get the right tile
        c = ctx.db.cursor()
        c.execute(
            (
                "SELECT tile_data FROM tiles WHERE "
                "zoom_level=? AND tile_column=? AND tile_row=?"
            ),
            (tile.zoom, tile.tile_x, tile.tile_y),
        )
        row = c.fetchone()
        if not row:
            tile.state = "done"
            return

        # no-file loading
        try:
            data = io.BytesIO(row[0])
        except Exception:
            # android issue, "buffer" does not have the buffer interface
            # ie row[0] buffer is not compatible with BytesIO on Android??
            data = io.BytesIO(bytes(row[0]))
        im = CoreImage(
            data,
            ext='png',
            filename="{}.{}.{}.png".format(tile.zoom, tile.tile_x, tile.tile_y),
        )

        if im is None:
            tile.state = "done"
            return

        return self._load_tile_done, (tile, im,)

    def _load_tile_done(self, tile, im):
        tile.texture = im.texture
        tile.state = "need-animation"

    def get_x(self, zoom, lon):
        if self.is_xy:
            return lon
        return super().get_x(zoom, lon)

    def get_y(self, zoom, lat):
        if self.is_xy:
            return lat
        return super().get_y(zoom, lat)

    def get_lon(self, zoom, x):
        if self.is_xy:
            return x
        return super().get_lon(zoom, x)

    def get_lat(self, zoom, y):
        if self.is_xy:
            return y
        return super().get_lat(zoom, y)



import sys
sys.path.append('..')
sys.path.append('.')
import time

class HttpDataLoader(DataLoader):
	def __init__(self):
		super(HttpDataLoader,self).__init__()
		self.hc = HttpClient()
		
	async def loadData(self,url,method='GET',params={},headers={}):
		try:
			resp = None
			if method=='GET':
				resp = await self.hc.get(url,params=params,headers=headers)
			else:
				resp = await self.hc.post(url,data=params,headers=headers)
				return resp
		except Exception as e:
			print('loadData(%s) Error ' % url,e)
			self.loadError(e)

if __name__ == '__main__':
	import sys
	from async_app import AsyncApp, wait_coro
	from kivy.uix.boxlayout import BoxLayout
	from kivy.uix.button import Button
	from kivy.uix.textinput import TextInput
	class MyApp(AsyncApp):
		def build(self):
			root = BoxLayout(orientation='vertical')
			btn = Button(text='get Remote data',size_hint_y=None,height=44)
			btn.bind(on_release=self.getData)
			root.add_widget(btn)
			self.txt = TextInput(multiline=True,readonly=True)
			root.add_widget(self.txt)
			return root
		def getData(self,btn):
			url = sys.argv[1] if len(sys.argv)>1 else 'https://www.baidu.com'
			hdl = HttpDataLoader()
			hdl.bind(on_loaded=self.showData)
			wait_coro(hdl.loadData,url)
			self.loop.call_later(0.001,hdl.loadData(url))

		def showData(self,instance,x):
			self.txt.text = x
	MyApp().run()



from .factory import *
from .date import *
from .time import *
from .int import *
from .float import *
from .amount import *
from .password import *
from .text import *
from .code import *
from .viewer import *




from kivy.uix.popup import Popup
from kivy.clock import Clock
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.textinput import TextInput
from ..widget_css import WidgetCSS

class Messager:
	def __init__(self, show_time=0, title=None, **kw):
		self.show_time = show_time
		self.title = title if title else 'Message'
		self.time_task = None
		self.w = Popup(content=BoxLayout(orientation='vertical'),      
						title=self.title,
						size_hint=(0.8,0.8), **kw)
		self.messager = TextInput(size=self.w.content.size,
				background_color=[0.9,0.9,0.9,1],
				foreground_color=[0.3,0.3,0.3,1],
				multiline=True,readonly=True)
		self.w.content.add_widget(self.messager)

	def set_time_task(self):
		if self.show_time:
			self.time_task = Clock.schedule_once(self.hide,
					self.show_time)
	def show_error(self,e):
		self.messager.text = '%s[error]%s\n' % (self.messager.text,str(e))
		self.w.open()
		self.set_time_task()

	def clear_messages(self):
		self.messager.text = ''

	def show_info(self,info):
		self.messager.text = '%s[info]%s\n' % (self.messager.text, info)
		self.w.open()
		self.set_time_task()

	def hide(self, *args):
		if self.time_task:
			self.time_task.cancel()
			self.time_task = None
		self.w.dismiss()

